
/**********************************************************************************
 * Package: GSM                                                                   *
 * Class  : DAlphaQED                                                             *
 *                                                                                *
 * Description:                                                                   *
 *      Auxiliary Theory of the Delta-Alpha-QED at the Z-pole                     *
 *      leptonic contribution to effective em coupling constant                   *
 *      up to three loops                                                         *
 *                                                                                *
 * Sources:                                                                       *
 *      (1) M. Steinhauser, Phys. Lett. B429 (1998) 158-161, hep-ph/9803313       *
 *      (2) J.H. Kuehn, M. Steinhauser, Phys. Lett. B437 (1998) 425-431           *
 *                                                     Hep-ph/9802241             *                                               
 *                                                                                *
 **********************************************************************************/
#include "TMath.h"

#include "Gfitter/GMath.h"
#include "Gfitter/GConstants.h"
#include "Gfitter/GTheory.h"
#include "Gfitter/GTheoryRef.h"
#include "Gfitter/GParameterRef.h"
#include "Gfitter/GReference.h"

#include "GSM/DAlphaQED.h"
#include "GSM/RunningAlphaQCD.h"

#include "GSM/ZMath.h"

#include "Riostream.h"

using namespace Gfitter;
using std::complex;

ClassImp(GSM::DAlphaQED)

GSM::DAlphaQED::DAlphaQED()
   : Gfitter::GAuxTheory(),
     m_isUpToDate_Update( kFALSE )     
{
   SetTheoryName( GetName() );
   SetExistDerivative( kFALSE );

   BookParameter( "MZ"        , &p_MZ );
   BookParameter( "mt"        , &p_mt );
   BookParameter( "DAlphaHad" , &p_DAlphaHad );
   //   BookParameter( "GSM::MWTree"   , &p_MW );
  
   BookTheory   ( "GSM::RunningAlphaQCD" , &t_AlphasRun );
}

void GSM::DAlphaQED::Initialise()
{
   // use calculation from Steinhauser and Kuehn
   m_Zfitter = kFALSE;

   // lepton masses
   m_mLep2[0] = GMath::IPow( GConstants::me()  , 2 );
   m_mLep2[1] = GMath::IPow( GConstants::mmu() , 2 );
   m_mLep2[2] = GMath::IPow( GConstants::mtau(), 2 );
  
   m_alp1pi   = GConstants::alphaQED()/TMath::Pi();
}

void GSM::DAlphaQED::UpdateLocalFlags( GReference& /* ref */ )
{
   m_isUpToDate_Update = kFALSE;
}

// Update parmeters and check if something has changed
void GSM::DAlphaQED::Update()
{
   if (m_isUpToDate_Update) return;
  
   // now, it is uptodate (I mean... it will be)
   m_isUpToDate_Update = kTRUE;
  
   Double_t MZ2   = p_MZ*p_MZ;
   Double_t mt2   = p_mt*p_mt;
  
   Double_t chQ21 = DAlphaLep(MZ2);
  
   // excluded top-contribution
   m_DAlphaQED    = chQ21 + p_DAlphaHad;  
  
   // included top-contribution
   // ZFitter method
   if ( m_Zfitter ) {
      chQ21         += 2.0*m_alp1pi*3.0*4/9.0*real( GSM::ZMath::I3(mt2, -MZ2 ,mt2, mt2) );
      m_DAlphaQEDtop = chQ21 + p_DAlphaHad + AlQCD()*m_alp1pi/4.0;  
   }
   // Steinhuser and Kuehn
   else  m_DAlphaQEDtop = chQ21 + p_DAlphaHad + DAlphatop() ;  
   
   SetUpToDate();
}

// leptonic contribution
// eq. (1) of hep-ph/9803313
// and sub-equation (5) - (10)
Double_t GSM::DAlphaQED::DAlphaLep( Double_t s )
{
   Double_t dalpha = 0;
   Double_t sum    = 0;
   Double_t Const  = 0;
    
   for (Int_t i = 0; i < 3; i++) {
      Double_t Ratio = m_mLep2[i]/s;
      Double_t LogR  = -TMath::Log( Ratio );
      
      // ZFitter method
      // comming from dizet_6_42.f (ZFitter) line 5667 
      if ( m_Zfitter ) {
         dalpha += ( m_alp1pi*( 1/9.0 + 1/3.0*(1.0 + 2.0*Ratio)
                                *real( GSM::ZMath::I0(m_mLep2[i], -s , m_mLep2[i], m_mLep2[i]) ) ) );
      }
      else {
         // PI(0) eq. (5) + some 1-loop corrections, 
         dalpha += -0.25*m_alp1pi*(20/9.0 - 4/3.0*LogR + 8.0*Ratio);
      }

      // PI(1) eq. (6)
      dalpha += ( m_alp1pi*m_alp1pi*((1.0 + 12.0*Ratio)/4.0*LogR + GMath::Zeta3() - 5/24.0) );
      // PI(2)A eq. (7)
      dalpha += 2.0*( GMath::IPow(m_alp1pi, 3)*(-0.25*(-121/48.0 + (-5.0 + 8.0*TMath::Log(2.0))*GMath::Zeta2()
                                                       -99/16.0*GMath::Zeta3() + 10.0*GMath::Zeta5() + 1/8.0*LogR)) );
    
      for (Int_t j = 0; j < 3; j++) {
         Double_t Logi = TMath::Log( s/m_mLep2[i] );
         Double_t Logj = TMath::Log( s/m_mLep2[j] );
            
         if (i == j) {
            // PI(2)F eq. (9)
            sum += ( GMath::IPow(m_alp1pi, 3)
                     *(-0.25*(- 307/216.0 - 8/3.0*GMath::Zeta2() + 545/144.0*GMath::Zeta3()
                              +(11/6.0 - 4/3.0*GMath::Zeta3())*Logi - 1/6.0*Logi*Logi + Const)) );
         }
         else if( i > j) {
            // PI(2)l eq. (8)
            sum += ( GMath::IPow(m_alp1pi, 3)
                     *(-0.25*( -116/27.0 + 4/3.0*GMath::Zeta2() + 38/9.0*GMath::Zeta3() 
                               + 14/9.0*Logi + (5/18.0 - 4/3.0*GMath::Zeta3())*Logj
                               + 1/6.0*Logi*Logi - 1/3.0*Logi*Logj + Const)) );
         }
         else {
            // PI(2)h eq. (10)
            sum += ( GMath::IPow(m_alp1pi, 3)
                     *(-0.25*(- 37/6.0 + 38/9.0*GMath::Zeta3() 
                              + (11/6.0 - 4/3.0*GMath::Zeta3())*Logj - 1/6.0*Logj*Logj + Const)) );
         }
      }
   }
  
   return dalpha + sum;   
}


// for DAlphaLept
Double_t GSM::DAlphaQED::AlQCD()
{
   // sw2 = 1 - MW^2/MZ^2
   // Attention: MW depends on DAlphaQED
   // => DAlphaQED -> MW -> DAlphaQED !!!
   // We use the measured value from the steering card
   // The obtained error is small because AlQCD is multiplied by a small value ( line 93 )
   
   if (TMath::IsNaN(p_mt)) m_logger << kFATAL << "<DAlphaQED::AlQCD> p_mt is NaN !" << GEndl;

   Double_t sw2   = 1.0 - p_MW->GetValue()*p_MW->GetValue()/(p_MZ*p_MZ);
   Double_t pvftz = -p_mt*p_mt/(p_MZ*p_MZ)*real( GSM::ZMath::PolVfi(-p_mt*p_mt/(p_MZ*p_MZ)) );
   Double_t alqcd = ( -GetAlphasRun().EvolveAlphas(p_mt)/(3.0*TMath::Pi()*sw2)
                      *(4.0*sw2*4/9.0*( pvftz - 45/4.0 )) );
   return alqcd;
}


// eq. (12) from hep-ph/9802241
Double_t GSM::DAlphaQED::DAlphatop()
{
   Double_t alphaspi = GetAlphasRun().EvolveAlphas(p_MZ)/TMath::Pi();
   Double_t L        = TMath::Log((p_MZ*p_MZ)/(p_mt*p_mt));

   Double_t atop = ( -4/45.0*m_alp1pi*(p_MZ*p_MZ)/(p_mt*p_mt)*
                     (1.0 + 5.062*alphaspi + (28.220 + 9.702*L)*alphaspi*alphaspi
                      + (p_MZ*p_MZ)/(p_mt*p_mt)*(0.1071 + 0.8315*alphaspi + (6.924 + 1.594*L)*alphaspi*alphaspi)) );
   return atop;
}
      
