/**********************************************************************************
 * Package: GSM                                                                   *
 * Class  : EW2Loop                                                               *
 *                                                                                *
 * Description:                                                                   *
 *      Electroweak two loop calculation for ew formfactors                       *
 *      of order O(alpha^2*mt^2/MW^2)                                             *
 *                                                                                *
 * Sources:                                                                       *
 *    (1) G. Degrassi, P. Gambino, Nucl. Phys. B567 (2000) 3-31, hep-ph/9905472   *
 *    (2) G. Degrassi, P. Gambino, A. Vicini,                                     *
                                  Phys.Lett. B383 (1996) 219-226, hep-ph/9603374  *
 *    (3) Degrassi, Bardin, The Standard Model in the Making, Oxford 1999         *
 *                                                                                *
 * The Code of this class has been ported to C++ from Fortran code authored       *
 *    by G. Degrassi and P. Gambino that is also used in the Fortran package      *
 *    ZFITTER.                                                                    *
 *                                                                                *
 **********************************************************************************/
#include "TMath.h"

#include "Gfitter/GMath.h"
#include "Gfitter/GConstants.h"
#include "Gfitter/GTheory.h"
#include "Gfitter/GTheoryRef.h"
#include "Gfitter/GParameterRef.h"
#include "Gfitter/GReference.h"
#include "Gfitter/GVariable.h"
#include "Gfitter/GStore.h"

#include "GSM/EW2Loop.h"
#include "GSM/ZMath.h"

using std::complex;

using namespace Gfitter;

GSM::EW2Loop::EW2Loop() 
   : Gfitter::GAuxTheory(),
     m_isUpToDate_Update( kFALSE )   
{ 
  
   SetTheoryName( GetName() );
   SetExistDerivative( kFALSE );

   const TString& logMH = gStore()->GetVariable( "GSMFlags::logMH" )->GetStringValue();
   m_logger << kINFO << "Using logMH: \"" << logMH << "\"" << GEndl;
   
   if      (logMH == "Yes" ) m_logMH  = kTRUE;
   else if (logMH == "No" )  m_logMH  = kFALSE;
   else {
      m_logger << kFATAL << "unknown value for \"GSMFlags::logMH\": \"" << logMH << "\""
               << ". Possible are: \"Yes\" and \"No\"\""
               << GEndl;
   }

   BookParameter( "MZ"      , & p_MZ );
   BookParameter( "mt"      , & p_mt );
   BookParameter( "GF"      , & p_GF );
   BookTheory   ( "GSM::WMass" , & t_MW );
   BookTheory   ( "GSM::MH" , & t_MH );
}

void GSM::EW2Loop::UpdateLocalFlags( GReference& /* ref */ )
{
   m_isUpToDate_Update = kFALSE;
}

void GSM::EW2Loop::Update()
{
   if (m_isUpToDate_Update) return;

   // now, it is uptodate (I mean... it will be)
   m_isUpToDate_Update = kTRUE;

   m_MZ        = p_MZ;  
   Double_t mt = p_mt;
   Double_t MW = GetMW();
   Double_t MH = GetMH().GetValue(); //p_MH;
   if( m_logMH ) MH = TMath::Exp( GetMH().GetValue() ); //p_MH );

   // protection by shifting of Higgs mass if it is = 2*mt or 2*mz
   if( TMath::Abs(MH - 2.0*mt) < 5e-5 ) MH = 2.0*mt - 1e-4;
   if( TMath::Abs(MH - 2.0*m_MZ) < 5e-5 ) MH = 2.0*m_MZ - 1e-4;

   m_Nc   = 3;

   m_c2   = GMath::IPow( MW/m_MZ ,2 );
   m_s2   = 1.0 - m_c2;

   // 't Hooft mass, here MZ
   m_mu   = m_MZ;

   m_sman = m_MZ*m_MZ;
   m_zt   = GMath::IPow( m_MZ/mt, 2 );
   m_ht   = GMath::IPow( MH/mt, 2 );
   m_xt   = p_GF/TMath::Sqrt(2.0)*mt*mt/(8.0*GMath::IPow( TMath::Pi() , 2 ));

   Double_t ec        = ECharge   ( m_zt, m_ht, m_s2, m_MZ, m_mu );
   Double_t Drho      = deltarho  ( m_zt, m_ht, m_s2, m_MZ, m_mu );
   Double_t DrhoOS    = deltarhoOS( m_zt, m_s2, m_MZ, m_mu );
  
   m_deltarw   = deltarw  ( m_zt, m_ht, m_s2, m_MZ, m_mu );
   m_deltarwOS = deltarwOS( m_zt, m_s2, m_MZ, m_mu );
 
   m_kappa     = kappa2L  ( m_zt, m_ht, m_c2, m_MZ, m_mu );
  
   m_deltaeta  = deltaeta ( m_zt, m_ht, m_c2, m_MZ, m_mu );
  
   m_Drho2L    = m_Nc*m_xt*m_xt*(Drho + 4.0*m_zt*m_c2*DrhoOS);
   m_Dkap2L    = m_Nc*m_xt*m_xt*(Drho + 4.0*m_zt*m_c2*DrhoOS);
   m_DrdRem    = ( m_Nc*GMath::IPow( (GConstants::alphaQED()/(4.0*TMath::Pi()*m_s2)), 2 )
                   *GMath::IPow( (mt*mt/(4.0*MW*MW)), 2 )
                   *(16.0*m_deltarw + 4.0*m_zt*m_c2*m_deltarwOS - 16.0*ec) );

   SetUpToDate();
} 



Double_t GSM::EW2Loop::deltarw(const Double_t& zt, const Double_t& ht, const Double_t& s2, const Double_t& mz, const Double_t& mu)
{
   Double_t c2 = 1.0 - s2;  
   Double_t mz2 = mz*mz;
   Double_t mu2 = mu*mu;
   Double_t drw = 0;
   Double_t pi = TMath::Pi();

   // [Aww (mw^2) - Aww(0)]/mw^2
   if( TMath::Sqrt(ht) < 0.3){
      // light Higgs case
      // see eq. (7a) of hep-ph/9603374 and ep. (A6) of hep-ph/9905472
      drw = ( 35.0*ht/288.0 - 5.0*ht*ht/(144.0*c2*zt) - 41.0*zt/96.0 - zt/(48.0*c2) + 325.0*c2*zt/144.0 
              - c2*TMath::Sqrt(ht)*pi*zt/3.0 - c2*pi*pi*zt/36.0 
              + ZMath::B0(c2*mz2, ht*mz2/zt,c2*mz2,mu2)*(- 5.0*ht/36.0 + 5.0*ht*ht/(144.0*c2*zt) + 5.0*c2*zt/12.0) 
              + ZMath::B0(c2*mz2, c2*mz2, mz2,mu2)*(5.0*zt/12.0  +  zt/(48.0*c2)  -  c2*zt/2.0)
              + (5.0*( -1.0)*ht/144.0 - 19.0*zt/12.0 - (425.0*c2*zt/144.0) + 25.0*zt/(16.0*s2) 
                 - 5.0*c2*c2*zt*zt/(16.0*(ht - c2*zt)))*TMath::Log(c2) + zt*GMath::IPow( TMath::Log(c2), 2 )/2.0 
              + c2*zt*GMath::IPow( TMath::Log(c2), 2 )/2.0 
              + zt*GMath::IPow( TMath::Log(c2), 2 )/(2.0*( -s2)) - 5.0*ht*TMath::Log(ht)/48.0  
              + 5.0*ht*ht*TMath::Log(ht)/(144.0*c2*zt) + 5.0*c2*zt*TMath::Log(ht)/16.0  
              + 5.0*c2*c2*GMath::IPow( zt, 2 )*TMath::Log(ht)/(16.0*(ht - c2*zt)) 
              + ( - 5.0*ht/36.0 + 5.0*ht*ht/(144.0*c2*zt) + 5.0*zt/12.0  
                  + zt/(48.0*c2)  - 7.0*c2*zt/3.0  +  2.0*zt*TMath::Log(c2)  
                  + 2.0*c2*zt*TMath::Log(c2) - 2.0*zt*TMath::Log(c2)/s2)
              *(TMath::Log(1.0/zt) + TMath::Log(mz2/mu2) ) 
              + ( -5.0*ht/144.0 + 5.0*zt/12.0 + zt/(48.0*c2) - 209.0*c2*zt/144.0  
                  - 5.0*c2*c2*zt*zt/(16.0*(ht  -  c2*zt)) + zt*TMath::Log(c2) + c2*zt*TMath::Log(c2) 
                  + zt*TMath::Log(c2)/( - s2))*TMath::Log(zt) );
   }
   else {
      // heavy Higgs case
      // see eq. (7b) of hep-ph/9603374 and ep. (A7) of hep-ph/9905472
      drw  = ( c2*(4.0  -  ht)*zt*ZMath::Lambda( -1.0 + 4.0/ht)/(12.0*ht) 
               + c2*(19.0 + 21.0*ht - 12.0*ht*ht - 31.0*GMath::IPow( ht, 3 ) 
                     + 9.0*GMath::IPow( ht, 4 ))*zt*TMath::DiLog(1.0 - ht)/(72.0*ht*ht)
               + ((1.0  +  69.0*c2  -  145.0*c2*c2)*zt*TMath::Log(c2))/(48.0*( - s2))
               + (zt*( - 228.0*c2*c2*ht + 228.0*GMath::IPow( c2, 3 )*ht + 18.0*ht*ht 
                       + 351.0*c2*ht*ht - 2226.0*c2*c2*ht*ht  + 1857.0*GMath::IPow( c2, 3 )*ht*ht 
                       - 36.0*c2*c2*GMath::IPow( ht, 3 ) + 36.0*GMath::IPow( c2*ht, 3 ) 
                       + 38.0*c2*c2*pi*pi - 38.0*GMath::IPow( c2 , 3 )*pi*pi + 42.0*c2*c2*ht*pi*pi 
                       - 42.0*GMath::IPow( c2, 3 )*ht*pi*pi 
                       + ZMath::B0(c2*mz2, c2*mz2, mz2,mu2)
                       *( -18.0*ht*ht - 342.0*c2*ht*ht + 792.0*c2*c2*ht*ht - 432.0*GMath::IPow( c2, 3 )*ht*ht) 
                       + ZMath::gdeg(ht)*(144.0*c2*c2*s2*std::pow(ht, 2.5 ) - 36.0*c2*c2*s2*std::pow( ht, 3.5 )) 
                       + 432.0*GMath::IPow( c2, 3 )*ht*ht*GMath::IPow( TMath::Log(c2), 2 )))/(864.0*c2*( - s2)*ht*ht) 
               + (c2*( - 31.0 + 2.0*ht - 9.0*ht*ht + 3.0*GMath::IPow( ht, 3 ))*zt*TMath::Log(ht))/(72.0*ht)  
               + ((1.0 + 20.0*c2 - 132.0*c2*c2)*zt/(48.0*c2) + (2.0*c2*c2*zt*TMath::Log(c2))/( - s2))
               *(TMath::Log(1.0/zt) + TMath::Log(mz2/mu2)) 
               + (zt*( -1.0 - 19.0*c2 + 93.0*c2*c2 - 73.0*GMath::IPow( c2, 3 ) 
                       + 48.0*GMath::IPow( c2, 3 )*TMath::Log(c2))*TMath::Log(zt))/(48.0*c2*( -s2)) 
               + (c2*( - 24.0  + 74.0*ht + 35.0*ht*ht - 49.0*GMath::IPow( ht, 3 ) 
                       + 9.0*GMath::IPow( ht, 4 ))*zt*ZMath::Phi(ht/4.0))/(144.0*ht*ht) );
   }
  
   //vertex contribution
  
   drw   += ( zt*( ( - 2.0*c2 + 5.0*c2*c2)*TMath::Log(c2)/(2.0*( -s2)) 
                   + (3.0*c2 - 3.0*c2*c2 - c2*c2*GMath::IPow( TMath::Log(c2), 2 ))/(2.0*( - s2))
                   + ( -c2 + c2*c2 - c2*c2*TMath::Log(c2))/( - s2)
                   *(TMath::Log(1.0/zt) + 2.0*TMath::Log(mz2/mu2)) ) );
  
   // see also m2tcor5_11.f line 328
   return drw;
}


Double_t GSM::EW2Loop::ECharge(const Double_t& zt, const Double_t& ht, const Double_t& s2, const Double_t& mz, const Double_t& mu)
{
   Double_t c2     = 1.0 - s2;  
   Double_t mz2    = mz*mz;
   Double_t mu2    = mu*mu;
   Double_t pi     = TMath::Pi();
   Double_t ec     = 0;
  
   // electric charge in units
   // Nc*(alfa0/(4.d0*pi*s2))^2/(zt*c2)^2    
  
   if( TMath::Sqrt(ht) < 0.3 ){
      // light Higgs case
      // see (8.71) of The Standad Model in the Making
      // (B5) of hep-ph/9905472 and m2tcor5_11.f line 431
      ec = ( -s2*c2*zt*
             ( (2.0 - TMath::Log(ht))/9.0 
               + (100.0 + 256.0*( -4.0 + 2.0*TMath::Sqrt(ht)*pi) + 96.0*TMath::Log(ht))/864.0 
               + 13.0/18.0*(TMath::Log(1.0/zt) + TMath::Log(mz2/mu2)) ) );
   }
  
   else{

      // heavy Higgs case
      // see (8.72) of The Standad Model in the Making,
      // (B5) of hep-ph/9905472 and m2tcor5_11.f line 440
      ec = ( -s2*c2*zt*( (100.0 - 25.0*ht + 96.0*TMath::Log(ht) + 48.0*ht*TMath::Log(ht) 
                          + 256.0*( - 4.0 + ht/2.0 - ( -4.0 + ht)*TMath::Sqrt(ht)*ZMath::gdeg(ht)/4.0 
                                    + (6.0 - ht)*ht*TMath::Log(ht)/4.0) 
                          - 64.0*ht*( - 4.0 + ht/2.0 - ( - 4.0 + ht)*TMath::Sqrt(ht)*ZMath::gdeg(ht)/4.0 
                                      + (6.0 - ht)*ht*TMath::Log(ht)/4.0))/(216.0*(4.0 - ht)) 
                         + 13.0/18.0*(TMath::Log(1.0/zt) + TMath::Log(mz2/mu2)) 
                         + 4.0*( -1.0 + ht)*ZMath::Phi(ht/4.0)/(9.0*( -4.0 + ht)*ht)) );
   }
  
   return ec;
}

Double_t GSM::EW2Loop::deltarho(const Double_t& zt, const Double_t& ht, const Double_t& s2, const Double_t& mz, const Double_t& mu)
{
   Double_t c2   = 1.0 - s2;  
   Double_t mz2  = mz*mz;
   Double_t mu2  = mu*mu;
   Double_t mt   = mz/TMath::Sqrt(zt);
   Double_t sht  = TMath::Sqrt(ht);
   Double_t pi   = TMath::Pi();
   Double_t dr    = 0;

   // [Aww (mw^2)/mw^2  -  Azz(mz^2)/mw^2] in units 
   // Nc*(alfa0/(4.0*pi*s2))^2/(4.0*zt*c2)^2.0
        
   if( ht < 0.25 ) {
      // light Higgs case
      // see eq. (10a) of hep-ph/0603374 and (A6) of hep-ph/9905472
      dr = ( 19.0 - (53.0*ht)/3.0 - 4.0*TMath::Sqrt(ht)*pi - 2.0*pi*pi + 2.0*ht*pi*pi 
             + 8.0*ht*ht/(9.0*zt) - 5.0*ht*ht/(9.0*c2*zt) + 3.0/2.0*std::pow( ht, 1.5 )*pi 
             + ZMath::B0(mz2, ht*mz2/zt, mz2,mu2)*(32.0*ht/9.0 - 8.0*ht*ht/(9.0*zt)) 
             + ZMath::B0(c2*mz2, ht*mz2/zt, c2*mz2,mu2)*( - 20.0*ht/9.0 + 5.0*ht*ht/(9.0*c2*zt)) 
             - 5.0*ht*TMath::Log(c2)/9.0 - 2.0*ht*TMath::Log(ht) - 8.0*ht*ht*TMath::Log(ht)/(9.0*zt) 
             + 5.0*ht*ht*TMath::Log(ht)/(9.0*c2*zt) 
             + ht*(5.0*ht - 8.0*c2*ht + 12.0*c2*zt)*(TMath::Log(1.0/zt) + TMath::Log(mz2/mu2))/(9.0*c2*zt) 
             + ht*TMath::Log(zt)/3.0  
             + zt*(( - 9.0 + 845.0*c2 + 427.0*c2*c2 - 366.0*GMath::IPow( c2, 3 ) + 136.0*c2*TMath::Sqrt(ht)*pi 
                     - 464.0*c2*c2*TMath::Sqrt(ht)*pi + 256.0*GMath::IPow( c2, 3 )*TMath::Sqrt(ht)*pi 
                     - 119.0*c2*pi*pi + 44.0*c2*c2*pi*pi - 396.0*c2*ZMath::B0(mz2, ht*mz2/zt, mz2, mu2) 
                     + ZMath::B0(mz2, c2*mz2, c2*mz2, mu2)*( - 18.0*c2 - 324.0*c2*c2 + 288.0*GMath::IPow( c2, 3 )) 
                     + 180.0*c2*c2*ZMath::B0(c2*mz2,ht*mz2/zt,c2*mz2,mu2) 
                     + ZMath::B0(c2*mz2, c2*mz2, mz2, mu2)*(9.0 + 180.0*c2 - 216.0*c2*c2))/(27.0*c2) 
                   + ( - 3.0 - 32.0*c2 - 48.0*c2*c2)*TMath::Log(c2)/9.0  
                   + (3.0 - 10.0*c2 - 328.0*c2*c2 + 56.0*GMath::IPow( c2, 3 ))
                   *(TMath::Log(1.0/zt) + TMath::Log(mz2/mu2))/(9.0*c2) 
                   + (3.0 - 11.0*c2 - 16.0*c2*c2 - 48.0*GMath::IPow( c2, 3 ))*TMath::Log(zt)/(9.0*c2)) 
             + 4.0*zt*ZMath::B0(mz2, ht*mz2/zt, mz2,mu2) );
   }
  
   else if(ht < 4.0) {
      // Interpolation between light and heavy Higgs case
      // m2tcor5_11.f line 524
      dr = ( - 15.642064 + 0.036381841*mt + TMath::Sqrt(sht)*(2.30111 - 0.013429*mt) 
             + sht*(0.0180877*mt - 9.95272) + ht*(5.68703 - 0.0156807*mt) 
             + ht*sht*(0.00536879*mt - 1.64687) + ht*ht*(0.185188 - 0.000646014*mt)
             + 8.0/9.0*zt*(4.0 - 26.0*c2 - 5.0*c2*c2)*TMath::Log(mz2/mu2) );
   }
   
   else {
      // heavy Higgs case
      // see eq. (10b) of hep-ph/0603374 and (A6) of hep-ph/9905472
      dr = ( ( 10800.0*c2*GMath::IPow( ht, 3 ) - 4428.0*c2*GMath::IPow( ht, 4 ) 
               + 432.0*c2*GMath::IPow( ht, 5 ) - 432.0*c2*GMath::IPow( ht*pi, 2 )  
               + 324.0*c2*GMath::IPow( ht, 3 )*pi*pi 
               - 54.0*c2*GMath::IPow( ht, 4 )*pi*pi)/(108.0*c2*(4.0 - ht)*GMath::IPow( ht, 3 ))
             + (( - 4.0 + ht)*TMath::Sqrt(ht)*ZMath::gdeg(ht))/2.0 
             + ((54.0*ht*ht - 135.0*GMath::IPow( ht, 3 ) + 108.0*GMath::IPow( ht, 4 ) 
                 - 27.0*GMath::IPow( ht, 5 ))* TMath::DiLog(1.0 - ht))/(9.0*GMath::IPow( ht, 3 ))  
             + (( - 5184.0*GMath::IPow( ht, 3 ) - 2592.0*GMath::IPow( ht, 4 ) 
                  + 2700*GMath::IPow( ht, 5 ) - 540*GMath::IPow( ht, 6 ) + 27.0*GMath::IPow( ht, 7 ))
                *TMath::Log(ht))/(54.0*GMath::IPow( ( -4.0 + ht), 2 )*GMath::IPow( ht, 3 ))
             + ( - 12960.0*GMath::IPow( ht, 2 ) + 14256.0*GMath::IPow( ht, 3 ) 
                 - 5994.0*GMath::IPow( ht, 4 ) + 1134.0*GMath::IPow( ht, 5 ) - 81.0*GMath::IPow( ht, 6 ))
             *ZMath::Phi(ht/4.0)/(54.0*GMath::IPow( ( -4.0  +  ht) ,2 )*ht*ht)
             + zt*( (3552.0*c2*c2*ht*ht - 144.0*GMath::IPow( ht, 3 ) + 10772.0*c2*GMath::IPow( ht, 3 ) 
                     + 6112.0*c2*c2*GMath::IPow( ht, 3 ) - 7392.0*GMath::IPow( c2*ht, 3 ) 
                     + 36.0*GMath::IPow( ht, 4 ) - 2134.0*c2*GMath::IPow( ht, 4 ) 
                     - 2742.0*c2*c2*GMath::IPow( ht, 4 ) - 136.0*c2*GMath::IPow( ht, 5 ) 
                     + 2872.0*GMath::IPow( c2, 3 )*GMath::IPow( ht, 4 ) + 248.0*c2*c2*GMath::IPow( ht, 5 )
                     - 256.0*GMath::IPow( c2, 3 )*GMath::IPow( ht, 5 ) - 592.0*c2*c2*ht*pi*pi 
                     + 244.0*c2*c2*GMath::IPow( ht, 2 )*pi*pi - 1904.0*c2*GMath::IPow( ht, 3 )*pi*pi 
                     + 872.0*c2*c2*GMath::IPow( ht, 3 )*pi*pi + 476.0*c2*GMath::IPow( ht, 4 )*pi*pi 
                     - 224.0*c2*c2*GMath::IPow( ht, 4 )*pi*pi)/(108.0*c2*(4.0  -  ht)*GMath::IPow( ht, 3 ))
                    - 2.0*(1.0 + 18.0*c2 - 16.0*c2*c2)*ZMath::B0(mz2,c2*mz2,c2*mz2,mu2)/3.0 
                    + (1.0 + 20.0*c2 - 24.0*c2*c2)*ZMath::B0(c2*mz2,c2*mz2,mz2,mu2)/(3.0*c2) 
                    + ( -34.0 + 116.0*c2 - 64.0*c2*c2)*( -4.0 + ht)*TMath::Sqrt(ht)*ZMath::gdeg(ht)/54 
                    + 40.0*( - s2)*(4.0  -  ht)*ZMath::Lambda( - 1.0  +  4.0/ht)/(3.0*ht) 
                    + (74.0*c2*ht - 12.0*c2*ht*ht - 24.0*c2*GMath::IPow( ht, 3 ) 
                       - 44.0*c2*GMath::IPow( ht, 4 ) + 18.0*c2*GMath::IPow( ht, 5 ))
                    *TMath::DiLog(1.0 - ht)/(9.0*GMath::IPow( ht, 3 )) 
                    + ( -3.0 - 42.0*c2 - 48.0*c2*c2)*TMath::Log(c2)/(9.0) 
                    + (23040.0*ht*ht - 30144.0*c2*ht*ht 
                       + GMath::IPow( ht, 3 )*( - 10884.0 + 17856.0*c2 + 1536.0*c2*c2)
                       + GMath::IPow( ht, 4 )*( 5094.0 - 14124.0*c2 + 6528.0*c2*c2)
                       + GMath::IPow( ht, 5 )*( - 2302.0 + 7094.0*c2 - 4288.0*c2*c2)
                       + GMath::IPow( ht, 6 )*( 476.0 - 1516.0*c2 + 896.0*c2*c2) 
                       + GMath::IPow( ht, 7 )*( -34.0 + 116.0*c2 - 64.0*c2*c2))
                    *TMath::Log(ht)/(54.0*GMath::IPow( ( -4.0 + ht), 2 )*GMath::IPow( ht, 3 )) 
                    - (3.0 + 122.0*c2 - 388.0*c2*c2 + 56.0*GMath::IPow( c2, 3 ))
                    *(TMath::Log(zt) -  TMath::Log(mz2/mu2))/(9.0*c2) 
                    + (3.0 + 5.0*c2 - 26.0*c2*c2 - 48.0*GMath::IPow( c2, 3 ))*TMath::Log(zt)/(9.0*c2) 
                    + (23040.0 - 23040.0*c2 - 25860.0*ht + 25344.0*c2*ht + 1536.0*c2*c2*ht 
                       + ht*ht*(10236.0 - 7872.0*c2  - 1920*c2*c2) 
                       + GMath::IPow( ht, 3 )*( -1890.0 - 2856.0*c2 + 384.0*c2*c2)
                       + GMath::IPow( ht, 4 )*( 144.0 + 2724.0*c2) 
                       - 672.0*c2*GMath::IPow( ht, 5 ) 
                       + 54.0*c2*GMath::IPow( ht, 6 ))
                    *ZMath::Phi(ht/4.0)/(54.0*GMath::IPow( ( -4.0 + ht), 2 )*ht*ht)) 
             + 4.0*zt*(1.0 - TMath::Log(ht) + TMath::Log(zt) - TMath::Log(mz2/mu2)) );
   }
   // see also m2tcor5_11.f line 455
  
   return dr;
}

// This functions computes the additional term to be added
// to the Msbar 2 loop expression of
// [Aww (mw^2)/mw^2  -  Azz(mz^2)/mw^2] 
// in units Nc*(alfa0/(4.0*pi*s2))^2/(4.0*zt*c2)
          
Double_t GSM::EW2Loop::deltarhoOS( const Double_t& zt, const Double_t& s2, const Double_t& mz, const Double_t& mu )
{
   Double_t c2          = 1.0 - s2;  
   // see eq. (8.56) of The Standard Model in the Making
   // and m2tcor5_11.f line 588
   Double_t drOS  =  - (2.0*(18.0 - 181.0*c2 + 166.0*c2*c2 - 216.0*GMath::IPow( c2, 3 ))/(27.0*c2)
                        + TMath::Sqrt( -1.0 + 4.0*c2)
                        *(2.0*(1.0 + 26.0*c2 + 24.0*c2*c2)
                          *TMath::ATan(1/TMath::Sqrt( -1.0 + 4.0*c2))/3.0  
                          - (1.0 + 11.0*c2 + 12.0*c2*c2)*TMath::ATan(TMath::Sqrt( -1.0 + 4.0*c2))/(3.0*c2*c2)) 
                        + TMath::Log(c2)/(6.0*c2*c2)*(1.0 + 9.0*c2 - 4.0*c2*c2 + 60.0*GMath::IPow( c2, 3 )) 
                        + ( -20.0 + 32.0*c2)*TMath::Log(zt)/9.0 
                        + TMath::Log(mz*mz/mu/mu)*(4.0/3.0 + 34.0/3.0*s2 - 1.0/c2) );
         
   return drOS;
}
 
// This subroutine computes the additional term to be added
// to the Msbar 2 loop expression of Deltar^w
// in units Nc*(alfa0/(4.0*pi*s2))^2/(4.0*zt*c2)

Double_t GSM::EW2Loop::deltarwOS( const Double_t& /* zt */, const Double_t& s2, 
                                  const Double_t& /* mz */, const Double_t& /* mu */ )
{
   Double_t c2     = 1.0 - s2;  
   // see eq. (8.56) of The Standard Model in the Making
   // and m2tcor5_11.f line 619
   Double_t drwOS  =  - ( 1.0/(3.0*c2) + 4.0 - (1.0 + 11.0*c2 + 12.0*c2*c2)/(3.0*c2*c2)
                          *TMath::Sqrt( -1.0 + 4.0*c2)*TMath::ATan(TMath::Sqrt( -1.0 + 4.0*c2))
                          - TMath::Log(c2)*(0.5/s2 + 3.0 - 1.5/c2 - 1.0/(6.0*c2*c2)) );
     
   return drwOS;
}


// This function contains the 2 loop expressions of
// - c/s A_{gamma Z}(mz^2/mz^2) and the vertex contribution
// in units Nc*(alfa0/(4.0*pi*s2))**2.0/(4*zt*c2)**2.0

Double_t GSM::EW2Loop::kappa2L( const Double_t& zt, const Double_t& ht, const Double_t& c2, const Double_t& mz, const Double_t& mu )
{
   Double_t mz2     = mz*mz;
   Double_t mu2     = mu*mu;
   Double_t hht     = ht/4.0;
  
   // eq. (B3) of hep-ph/9905472
   Double_t kapcur  = zt*( - 8.0*c2*c2*TMath::Log(c2)/3.0 
                           + 4.0*(30.0*c2 - 48.0*c2*c2 
                                  + 135.0*c2*ht - 216.0*c2*c2*ht - 50.0*c2*ht*ht + 80.0*c2*c2*ht*ht 
                                  + 5.0*c2*GMath::IPow( ht, 3 ) - 8.0*c2*c2*GMath::IPow( ht, 3 ))
                           *TMath::Log(ht)/(27.0*( -4.0 + ht))  
                           + (251.0*c2 - 462.0*c2*c2 - 40.0*c2*ht + 64.0*c2*c2*ht 
                              + (18.0*c2 + 144.0*c2*c2)*ZMath::B0(mz2,mz2*c2,mz2*c2,mu2) 
                              + (20.0*c2 - 32.0*c2*c2)*(ht - 4.0)*TMath::Sqrt(ht)*ZMath::gdeg(ht) 
                              - (186.0*c2 - 240.0*c2*c2)*( - TMath::Log(zt) + TMath::Log(mz2/mu2)))/27.0
                           + 8.0*(c2 - 3.0*c2*c2)*TMath::Log(zt)/9.0 
                           + 8.0*ZMath::Phi(hht)*(5.0*c2 - 8.0*c2*c2)/(9.0*( -4.0 + ht)*ht) 
                           - 8.0*ZMath::Phi(hht)*(5.0*c2 - 8.0*c2*c2)/(9.0*( -4.0 +  ht)) );
  
   return kapcur;
}

// This function contains the additional piece to be
// added to the MSbar two - loop contribution to
// c/s A_{gamma Z}(mz^2/mz^2 and the vertex part
// in units Nc*(alfa0/(4.0*pi*s2))^2.0/(4*zt)
     
// i3qf = T3*ch
Double_t GSM::EW2Loop::kappa2LOS( const Double_t& zt, const Double_t& c2w, const Double_t& mz, 
                                  const Double_t& mu, const Double_t& sman, const Double_t& i3qf)
{        
   Double_t s2w     = 1.0 - c2w;
   Double_t mz2     = mz*mz;
   Double_t mu2     = mu*mu;
  
   Double_t kapOS = 0;

   // eq. (8.65) of The Standard Model in the Making
   // and eq. (B8) of hep-ph/9905472
   // top
   kapOS  =  -32.0/3.0*(ZMath::J1(1.0/zt,sman/mz2) + TMath::Log(sman/mu2)/6.0);
   // light quarks
   kapOS += 44.0/9.0*( - TMath::Log(sman/mu2) + 5.0/3.0);
   // leptons
   kapOS += 4.0*( -TMath::Log(sman/mu2) + 5.0/3.0);
   // bosons
   kapOS += -( 4.0*c2w*mz2/sman*TMath::Log(c2w) + 4.0*c2w*mz2/sman*TMath::Log(mz2/mu2) 
               + (3.0*sman/mz2 + 4.0*c2w*mz2/sman)*ZMath::B0(sman,mz2*c2w,mz2*c2w,mu2));
   // vertex contribution
   kapOS += ( 2.0*(ZMath::ug(c2w*mz2/sman) + 0.25) + (i3qf - 1.0)
              *(ZMath::revf(mz2*c2w/sman) + 3.0/2.0/c2w/c2w*ZMath::revf(mz2/sman))
              - 4.0*s2w/c2w*(i3qf*i3qf - 1.0)*ZMath::revf(mz2/sman) 
              + (5.0 - 16.0*s2w + 8.0*s2w*s2w)/4.0/c2w/c2w*ZMath::revf(mz2/sman)
              + 2.0*TMath::Log(c2w) + 2.0*TMath::Log(mz2/mu2) );
  
   return kapOS;
}


// This subroutine contains the 2 loop expressions of
// Re [Azz (q^2)  -  Azz(mz^2)]/(q^2  -  mz^2) |_mz^2
// in units Nc*(alfa0/(4.0*pi*s2))**2.0/(zt*c2)

Double_t GSM::EW2Loop::deltaeta( const Double_t& zt, const Double_t& ht, const Double_t& c2, const Double_t& mz, const Double_t& mu)
{
   Double_t mz2 = mz*mz;
   Double_t mu2 = mu*mu;
   Double_t pi  = TMath::Pi();

   Double_t deta = 0;
  
   if(TMath::Sqrt(ht) < 0.57) {
      // light Higgs case
      // eq. (B1) of hep-ph/9905472      
      deta = ( (( -17.0 + 40.0*c2 - 32.0*(c2*c2))*( -4.0 + 2.0*TMath::Sqrt(ht)*pi))/(108.0*c2)
               + (( -24.0 + 96.0*c2)*ht*ht*ht + (144.0 - 576.0*c2)*ht*ht*zt 
                  + ( -313.0 + 1345.0*c2 + 349.0*c2*c2 - 292.0*GMath::IPow( c2, 3 ))*ht*zt*zt 
                  + (196.0 - 1156.0*c2 - 1396.0*c2*c2 + 1168.0*GMath::IPow( c2, 3 ))*zt*zt*zt)
               /(216.0*c2*( -1.0 + 4.0*c2)*(ht - 4.0*zt)*zt*zt)  
               + (( -2.0*ht*ht*ht + 13.0*ht*ht*zt - 32.0*ht*zt*zt + 36.0*zt*zt*zt)
                  *ZMath::B0(mz2, ht*mz2/zt, mz2, mu2))/(18.0*c2*(ht - 4.0*zt)*zt*zt)  
               + (( -1.0 + 4.0*c2 - 44.0*c2*c2 + 32.0*GMath::IPow( c2, 3 ))
                  *ZMath::B0(mz2, c2*mz2, c2*mz2,mu2))/( -24.0*c2 + 96.0*(c2*c2))  
               +  (( -1.0 - 18.0*c2 + 16.0*c2*c2)*TMath::Log(c2))/(2.0*( -6.0 + 24.0*c2))
               +  (( -2.0*ht*ht*ht + 11.0*ht*ht*zt - 24.0*ht*zt*zt + 24.0*zt*zt*zt)
                   *TMath::Log(ht))/(18.0*c2*(ht - 4.0*zt)*zt*zt)  
               +  (((8.0 - 32.0*c2)*ht*ht*ht - 52.0*ht*ht*zt + 208.0*c2*ht*ht*zt + 159.0*ht*zt*zt  
                    - 704.0*c2*ht*zt*zt + 192.0*c2*c2*ht*zt*zt - 112.0*GMath::IPow( c2, 3 )*ht*zt*zt  
                    + ( -268.0 + 1344.0*c2 - 768.0*(c2*c2) + 448.0*GMath::IPow( c2, 3 ))*zt*zt*zt)
                   *(TMath::Log(1.0/zt) + TMath::Log(mz2/mu2)))
               /(72.0*c2*( -1.0 + 4.0*c2)*(ht - 4.0*zt)*zt*zt) 
               + (( -4.0*ht*ht +  16.0*c2*ht*ht + 20.0*ht*zt - 79.0*c2*ht*zt - 70.0*c2*c2*ht*zt 
                    + 48.0*GMath::IPow( c2, 3 )*ht*zt - 40.0*zt*zt + 156.0*c2*zt*zt + 280.0*c2*c2*zt*zt  
                    - 192.0*GMath::IPow( c2, 3 )*zt*zt)*TMath::Log(zt))/(36.0*c2*( -1.0 + 4.0*c2)*(ht - 4.0*zt)*zt) );
   }
   
   else {
      // heavy Higgs case
      // eq. (B2) of hep-ph/9905472      
      deta = ( ( - 1152.0 + 4608.0*c2 + 50.0*ht + 2248.0*c2*ht + 976.0*c2*c2*ht  
                 - 1600.0*GMath::IPow( c2, 3 )*ht + 67.0*ht*ht - 880.0*c2*ht*ht 
                 - 244.0*c2*c2*ht*ht + 400.0*GMath::IPow( c2, 3 )*ht*ht)/(864.0*c2*(1.0 - 4.0*c2)*( -4.0 + ht)*ht) 
               + (( - 1.0  +  4.0*c2  -  44.0*c2*c2  +  32.0*GMath::IPow( c2, 3 ))
                  *ZMath::B0(mz2,c2*mz2,c2*mz2,mu2))/( -24.0*c2 + 96.0*(c2*c2))  
               + (4.0/3.0 - 4.0/ht - (1.0 - 4.0/ht)*ZMath::Lambda( -1.0 + 4.0/ht))/(12.0*c2)  
               + (( - 1.0 - 18.0*c2 + 16.0*c2*c2)*TMath::Log(c2))/(2.0*( -6.0 + 24.0*c2)) 
               + (( - 384.0 - 202.0*ht + 320.0*c2*ht - 256.0*c2*c2*ht + 55.0*ht*ht + 80.0*c2*ht*ht  
                    - 64.0*c2*c2*ht*ht + 3.0*ht*ht*ht - 40.0*c2*ht*ht*ht + 32.0*(c2*c2)*ht*ht*ht)
                  *TMath::Log(ht))/(144.0*c2*GMath::IPow( ( -4.0 + ht), 2 )*ht)  
               + (( - 17.0 + 40.0*c2 - 32.0*(c2*c2))
                  *( - 4.0 + ht/2.0 + (1.0 - ht/4.0)*TMath::Sqrt(ht)*ZMath::gdeg(ht) 
                     + ((6.0 - ht)*ht*TMath::Log(ht))/4.0))/(108.0*c2)
               + (( - 31.0 + 192.0*c2 - 192.0*(c2*c2) + 112.0*GMath::IPow( c2, 3 ))
                  *(TMath::Log(1.0/zt) + TMath::Log(mz2/mu2)))/(72.0*c2  -  288.0*(c2*c2))  
               + (( 2.0 - 7.0*c2 - 70.0*c2*c2 + 48.0*GMath::IPow( c2, 3 ))*TMath::Log(zt))/( - 36.0*c2 + 144.0*(c2*c2)) 
               + (( - 384.0 - 10.0*ht + 320.0*c2*ht - 256.0*(c2*c2)*ht + 238.0*ht*ht - 400.0*c2*ht*ht  
                    + 320.0*c2*c2*ht*ht - 63.0*ht*ht*ht + 80.0*c2*ht*ht*ht - 64.0*(c2*c2)*ht*ht*ht  
                    + 3.0*GMath::IPow( ht, 4 ))*ZMath::Phi(ht/4.0))/(144.0*c2*GMath::IPow( ( -4.0 + ht), 2 )*ht*ht) );
   }
   // Information to this value also in 
   // The Standad Model in the Making page 432 
  
   return deta;
}
 

// This subroutine contains the additional piece to be
// added to the MSbar two - loop contribution to
// Re [Azz (q^2)  -  Azz(mz^2)]/(q^2  -  mz^2) |_mz^2
// in units Nc*(alfa0/(4.0*pi*s2))^2/(4.0*zt*c2)

Double_t GSM::EW2Loop::deltaetaOS( const Double_t& zt, const Double_t& ht, const Double_t& s2, 
                                   const Double_t& mz, const Double_t& mu, const Double_t& sman, const Double_t& i3qf)
{  
   Double_t mz2    = mz*mz; 
   Double_t mu2    = mu*mu;
   Double_t zs2    = 1.0/zt ;       
   Double_t c2     = 1.0 - s2;
   Double_t detaOS = 0;

   // Information to this value in hep-ph/9905472 particulary eq. (B6)
   // and in m2tcor5_11.f line 484


   // Azz
   detaOS = ( (197.0 - 1378.0*c2 + 1064.0*c2*c2)/(27.0*( -1.0 + 4.0*c2)) 
              + (1.0 + 16.0*c2 - 20.0*(c2*c2) + 48.0*GMath::IPow( c2, 3 ))
              *ZMath::B0(mz2, c2*mz2, c2*mz2, mu2)/( -3.0 + 12.0*c2)  
              + (2.0*c2*(1.0 + 26.0*c2 + 24.0*(c2*c2))*TMath::Log(c2))/( -3.0 + 12.0*c2)  
              + (2.0*( -20.0 + 113.0*c2 - 102.0*(c2*c2) + 24.0*GMath::IPow( c2,3 ))
                 *(TMath::Log(zs2) + TMath::Log(mz2/mu2)))/( -3.0 + 12.0*c2)  
              + (2.0*( -50.0 + 283.0*c2 - 242.0*c2*c2 + 72.0*GMath::IPow( c2, 3 ))
                 *TMath::Log(zt))/( -9.0 + 36.0*c2) );  

   // Vertex contribution 
   detaOS += ( 4.0*c2*(TMath::Log(c2) + TMath::Log(mz2/mu2)) + 4.0*c2*(ZMath::ug(mz2*c2/sman) + 0.25)  
               - 2.0*c2*(1.0 - i3qf)*ZMath::revf(mz2*c2/sman) 
               - (1.0 - 6.0*i3qf + 12.0*(1.0 - c2*c2)*i3qf*i3qf)/(2.0*c2)
               *ZMath::revf(mz2/sman) ); 
  
   // Contribution of eta due to overall shift
  
   // hadronic contribution
   Double_t di = 4.0*zs2 - 1.0;
   detaOS += ( (6.0*(zs2 - 4.0*zs2*zs2/di*ZMath::Lambda(di) + 1.0/6.0 + ZMath::J1(zs2,1.0))
                *(16.0*s2*s2/9.0 - 4.0*s2/3.0 + 0.5) - 1.5*zs2  
                + 6.0*zs2*zs2/di*ZMath::Lambda(di) - 2.0/3.0*(2.5 - 14.0/3.0*s2 + 44.0*s2*s2/9.0)   
                + 3.0*(20.0/9.0*s2*s2 - 2.0*s2 + 1.0)*TMath::Log(mz2/mu2))/c2 );
  
   // leptonic contribution
   detaOS += (1.0 - 2.0*s2 + 4.0*s2*s2)/c2*(TMath::Log(mz2/mu2) - 2.0/3.0);
  
   // bosonic contribution
   Double_t a2 = TMath::Sqrt(4.0*c2 - 1.0);
   a2  = a2*TMath::ATan(1.0/a2);
   a2  = a2*( 1.0/6.0/c2 - 2.0/3.0 + c2/c2/3.0 - 4.0/3.0*c2
              + 6.0*c2*c2 - 6.0*c2 + c2/c2*(18.0*c2*c2 - 8.0*c2*GMath::IPow( s2, 2 ))/(4.0*c2 - 1.0) );

   a2 += ZMath::hbar(ht/zt)/2.0/c2;
   a2 += ( TMath::Log(c2)*(1.0/12.0/c2 - 1.0/3.0 - 3.0*c2) - 4.0*c2*c2
           + 108.0/36.0*c2 - 8.0*c2/3.0 + 5.0/9.0 - 5.0/36.0/c2
           + 5.0/3.0 - 1.0/6.0*(19.0*c2 - s2*s2/c2)*TMath::Log(mz2/mu2) );
  
   detaOS += a2;
  
   return detaOS;
}
