/**********************************************************************************
 * Package: GSM                                                                   *
 * Class  : FermionPart                                                    *
 *                                                                                *
 * Description:                                                                   *
 *      Auxiliary Theory for fermionic part of self energies                      *
 *      one loop core of ZFitter option                                           *
 *                                                                                *
 **********************************************************************************/
#ifndef GSM_FermionPart
#define GSM_FermionPart

#include <complex>
#include "Gfitter/GAuxTheory.h"
#include "GSM/QPoleMass.h"
#include "GSM/WMass.h"

class Gfitter::GReference;
class Gfitter::GParameterRef;
class Gfitter::GTheoryRef;

namespace GSM {
      
   class FermionPart : public Gfitter::GAuxTheory {
          
   public:
     
      FermionPart();
      virtual ~FermionPart() {}

      // accessors, downcast for using them
      std::complex<Double_t> GetWfAt0();
      std::complex<Double_t> GetZfAt0();
      std::complex<Double_t> GetWfAtMW();
      std::complex<Double_t> GetZfAtMZ();
      std::complex<Double_t> GetWfFAtMW();
      std::complex<Double_t> GetZfFAtMZ();
      std::complex<Double_t> GetMfPhoZAtMZ();
      
      //not used
      void Initialise() {}
      
   protected:

      // update of flags in case of parameter change
      void UpdateLocalFlags( Gfitter::GReference& ref );
   
      // update parameters
      void Update();
 
      // reference to W mass
      Double_t GetMW() const { return ((GSM::WMass&)*t_MW).GetValue(); }      
      Gfitter::GTheoryRef t_MW; 
     
      // reference to pole quark masses
      GSM::QPoleMass& GetQPoleMass() const { return (GSM::QPoleMass&)*t_QPoleMass; }      
      Gfitter::GTheoryRef    t_QPoleMass;  
      
      // Data-Card members
      Gfitter::GParameterRef p_MZ; 
      Gfitter::GParameterRef p_mt; 
      

      // updated variables
      Double_t m_MW2;
      Double_t m_MZ2;
     
      Double_t m_R;
      Double_t m_R1;
      Double_t m_R12;

      // member variables
      // self energies etc.
      std::complex<Double_t> m_WfAt0;
      std::complex<Double_t> m_ZfAt0;
      std::complex<Double_t> m_WfAtMW;
      std::complex<Double_t> m_ZfAtMZ;
      std::complex<Double_t> m_ZfFAtMZ;
      std::complex<Double_t> m_WfFAtMW;
      std::complex<Double_t> m_MfPhoZAtMZ;

      // lepton masses
      Double_t m_ml[2][3];
      // quark masses
      Double_t m_mq[2][3];
      // quark charges
      Double_t m_chq[2];

      // functions
      std::complex<Double_t> MfphoZ( const Double_t& Q2 ) const;
      std::complex<Double_t> DerivativeWF( const Double_t& Q2 ) const;
      std::complex<Double_t> DerivativeZF( const Double_t& Q2 ) const;

   private:
    
      // boolean flags to track recomputation
      Bool_t m_isUpToDate_Update; 

      ClassDef(FermionPart,0) 
   };
}
//--------------------- inline functions for fast access --------------------

// W boson self energy
inline std::complex<Double_t> GSM::FermionPart::GetWfAt0()
{ 
   Update();
   return m_WfAt0;
}

inline std::complex<Double_t> GSM::FermionPart::GetWfAtMW()
{
   Update();
   return m_WfAtMW; 
}

// Z boson self energy
inline std::complex<Double_t> GSM::FermionPart::GetZfAt0()
{
   Update();
   return m_ZfAt0; 
}

inline std::complex<Double_t> GSM::FermionPart::GetZfAtMZ()
{ 
   Update();
   return m_ZfAtMZ;
}

// derivatives of self energies / F = finite
inline std::complex<Double_t> GSM::FermionPart::GetZfFAtMZ()
{
   Update();
   return m_ZfFAtMZ; 
}

inline std::complex<Double_t> GSM::FermionPart::GetWfFAtMW()
{ 
   Update();
   return m_WfFAtMW;
}

// photon Z mixing function
inline std::complex<Double_t> GSM::FermionPart::GetMfPhoZAtMZ()
{
   Update();
   return m_MfPhoZAtMZ;
}

#endif
