/**********************************************************************************
 * Package: GSM                                                                   *
 * Class  : GammaZlept                                                            *
 *                                                                                *
 * Description:                                                                   *
 *      Prediction for partial width of Z0->ll                                    *
 *      simple formulae                                                           *
 *                                                                                *
 * Source:                                                                        *
 *     A. Ferroglia, G. Ossola, M. Passera, A. Sirlin,                            *
 *        Phys. Rev. D65 (2002) 113002, hep-ph/0203224.                           *
 *                                                                                *
 **********************************************************************************/
#include "TMath.h"

#include "Gfitter/GTheory.h"
#include "Gfitter/GTheoryRef.h"
#include "Gfitter/GParameterRef.h"
#include "Gfitter/GReference.h"
#include "Gfitter/GVariable.h"
#include "Gfitter/GStore.h"

#include "GSM/AlphaQCDAtQ.h"
#include "GSM/GammaZlept.h"
#include "GSM/MH.h"

using namespace Gfitter;

ClassImp(GSM::GammaZlept)

GSM::GammaZlept::GammaZlept()
   : Gfitter::GTheory()
{
   SetTheoryName( GetName() );
   SetExistDerivative( kFALSE );
  
   // theory choice
   const TString& omstype = gStore()->GetVariable( "GSMFlags::OMSType" )->GetStringValue();
   m_logger << kINFO << "Using OMs type: \"" << omstype << "\"" << GEndl;
   
   if      (omstype == "OMS1" ){
      m_OMSType  = OMS1;
   }
   else if (omstype == "OMS1_2"){
      m_OMSType  = OMS1_2;
   }
   else if (omstype == "OMS2"){
      m_OMSType  = OMS2;
   }
   else {
      m_logger << kFATAL << "unknown value for \"GSMFlags::OMStype\": \"" << omstype << "\""
               << ". Possible are: \"OMS1\", \"OMS1-2\" \"OMS2\""
               << GEndl;     
   }
  
   const TString& logMH = gStore()->GetVariable( "GSMFlags::logMH" )->GetStringValue();
   m_logger << kINFO << "Using logMH: \"" << logMH << "\"" << GEndl;
   
   if      (logMH == "Yes" ) m_logMH  = kTRUE;
   else if (logMH == "No" )  m_logMH  = kFALSE;
   else {
      m_logger << kFATAL << "unknown value for \"GSMFlags::logMH\": \"" << logMH << "\""
               << ". Possible are: \"Yes\" and \"No\"\""
               << GEndl;
   }
   
   BookParameter( "mt",             &p_mt );   
   BookParameter( "MZ",             &p_MZ );   
   BookParameter( "DAlphaHad",      &p_DAlphaHad );    
   
   BookTheory   ( "GSM::RunningAlphaQCD", &t_AlphasRun );
   BookTheory   ( "GSM::MH",              &t_MH );
}

// coefficients, see table 3 of hep-ph/0203224
// we use the OS1 coefficients 
void GSM::GammaZlept::Initialise()
{
   switch(m_OMSType){
   case OMS1:
   case OMS1_2: // here no intermediate step 
      m_gZ0 = 84.0174;
      m_g1  = 0.04472;
      m_g2  = 0.476;
      m_g3  = 0.804;
      m_g4  = 0.113;
      m_g5  = 0.03673;
      break;
   case OMS2:
      m_gZ0 = 84.0152;
      m_g1  = 0.04568;
      m_g2  = 0.476;
      m_g3  = 0.801;
      m_g4  = 0.112;
      m_g5  = 0.03683;
      break;
   }
}

// eq. (23) of hep-ph/0203224
Double_t GSM::GammaZlept::GetTheoryPrediction()
{
   Double_t MH        = GetMH().GetValue(); //p_MH;
   if( m_logMH ) MH = TMath::Exp( GetMH().GetValue() ); //p_MH );

   Double_t A1  = TMath::Log(MH/100.0);
   Double_t A2  = (p_DAlphaHad/0.02761) - 1.0;
   Double_t A3  = ((p_mt/174.3)*(p_mt/174.3) - 1.0);
   Double_t A4  = (GetAlphasRun().EvolveAlphas(p_MZ)/0.118) - 1.0;   
     
   Double_t gzlep  = ( m_gZ0 - m_g1*A1 - m_g5*A1*A1 - m_g2*A2 + m_g3*A3 - m_g4*A4 );

   return gzlep;
}
