/**********************************************************************************
 * Package: GSM                                                                   *
 * Class  : RadiatorFunctions                                                     *
 *                                                                                *
 * Description:                                                                   *
 *      Auxiliary Theory computes QCD radiator functions                          *
 *      for Z and W decay                                                         *
 *                                                                                *
 * Papers:                                                                        *
 *      - Degrassi, Bardin, The Standard Model in the Making, Oxford 1999         *
 *      - Bardin et al, CPC 133 (2001) 229-395, hep-ph/9908433                    *
 *      - Baikov et al., PR Lett. 101(2008) 012002, arXiv:0709.1983               *
 *      - Bardin et al, ZFitter package dizet6_42.f                               *
 *                                                                                *
 * This class also contains code lines ported to C++ from the Fortran package     *
 * ZFITTER                                                                        *
 *                                                                                *
 **********************************************************************************/
#include "TMath.h"
#include "Riostream.h"

#include "Gfitter/GMath.h"
#include "Gfitter/GStore.h"
#include "Gfitter/GConstants.h"
#include "Gfitter/GParameterRef.h"
#include "Gfitter/GTheoryRef.h"
#include "Gfitter/GVariable.h"

#include "GSM/RadiatorFunctions.h"

#include "GSM/QMassRunning.h"
#include "GSM/DAlphaQED.h"
#include "GSM/AlphaQCDAtQ.h"

#include <iomanip>

using namespace Gfitter;

ClassImp(GSM::RadiatorFunctions)

GSM::RadiatorFunctions::RadiatorFunctions()
   : Gfitter::GAuxTheory(),
   m_useNLOAlphasOnly ( kFALSE ),
   m_isUpToDate_Update( kFALSE )   
{
   SetTheoryName( GetName() );
   SetExistDerivative( kFALSE );

   BookParameter( "MZ",                        &p_MZ );   
   BookParameter( "mt",                        &p_mt );   
   BookParameter( "DeltaAlphasTheoC05_Scale",  &p_DeltaAlphasTheoC05_Scale );
   BookParameter( "DeltaAlphasTheoCMt3_Scale", &p_DeltaAlphasTheoCMt3_Scale );
 
   BookTheory   ( "GSM::RunningAlphaQCD",      &t_AlphasRun );
   BookTheory   ( "GSM::DAlphaQED",            &t_DAlphaQED );
   BookTheory   ( "GSM::QMassRunning",         &t_qMassRun );
}

void GSM::RadiatorFunctions::Initialise()
{
   if (gStore()->ExistVariable( "GSMFlags::UseNLOAlphasOnlyInRadFun" )) {
      m_useNLOAlphasOnly = gStore()->GetVariable( "GSMFlags::UseNLOAlphasOnlyInRadFun" )->GetBoolValue();
   }

   Double_t zeta2 = GMath::Zeta2();
   Double_t zeta3 = GMath::Zeta3();
   Double_t zeta4 = GMath::Zeta4();
   Double_t zeta5 = GMath::Zeta5();

   // five active falvors
   m_nf1   = 5;
   
   // see for coefficients The Standard Model in the Making page 504
   // and hep-ph/9908433 page 100-102
  
   // massless non-singlet corrections   
   m_C01   = 1.0;
   m_C02   = 365/24.0 - 11*zeta3 + (-11/12.0 + 2/3.0*zeta3)*m_nf1;
   m_C03   = ( 87029/288.0 - 121/8.0*zeta2 - 1103/4.0*zeta3 + 275/6.0*zeta5 
               + (-7847/216.0 + 11/6.0*zeta2 + 262/9.0*zeta3 - 25/9.0*zeta5)*m_nf1 
               + (151/162.0 - 1/18.0*zeta2 - 19/27.0*zeta3)*m_nf1*m_nf1 );
   // NNNLO calculation from arXiv:0709.1983
   m_C04   = -156.61 + 18.77*m_nf1 -0.7974*m_nf1*m_nf1 + 0.0215*m_nf1*m_nf1*m_nf1;
   
   // rough estimate from geometric series: C05 = C04*(C04/C03)
   m_C05   = -68.078*(1 - p_DeltaAlphasTheoC05_Scale);
   
   // Quadratic massive corrections
   // only running mc and mb => sqrt(s) > 15 GeV
   // light quarks
   m_CL1   = 0;
   m_CL2   = 0;
   m_CL3   = -80.0 + 60*zeta3 + (32/9.0 - 8/3.0*zeta3)*m_nf1;
   
   // only heavy quarks
   // vector part
   m_CV1   = 12.0;
   m_CV2   = 253/2.0 - 13/3.0*m_nf1;
   m_CV3   = ( 2522.0 - 855/2.0*zeta2 + 310/3.0*zeta3 - 5225/6.0*zeta5
               + (-4942/27.0 + 34*zeta2 - 394/27.0*zeta3 + 1045/27.0*zeta5)*m_nf1 
               + (125/54.0 - 2/3.0*zeta2)*m_nf1*m_nf1 );
     
   // axial part
   m_CA0   = -6.0;
   m_CA1   = -22.0;
   m_CA2   = -8221/24.0 + 57*zeta2 + 117*zeta3 + (151/12.0 - 2*zeta2 - 4*zeta3)*m_nf1;
   m_CA3   = ( - 4544045/864.0 + 1340*zeta2 + 118915/36.0*zeta3 - 1270*zeta5
               + (71621/162.0 - 209/2.0*zeta2 - 216*zeta3 + 5*zeta4 + 55*zeta5)*m_nf1
               + (-13171/1944.0 + 16/9.0*zeta2 + 26/9.0*zeta3)*m_nf1*m_nf1 );

   // Quartic massive corrections
   m_C42   = 13/3.0 - 4.0*GMath::Zeta3();

   m_CV40  = -6.;
   m_CV41  = -22.;
   // first term -3029/12. replaced! 
   // (taken from dizet6_42.f line 5120)
   m_CV42  = -3173/12.0 + 162.*GMath::Zeta2() + 112.0*GMath::Zeta3() + 
      (143/18.0 - 4.*GMath::Zeta2() - 8/3.0*GMath::Zeta3())*m_nf1;

   m_CVL42 = -11/2. + 1/3.*m_nf1;

   m_CA40  = 6.;
   m_CA41  = 10.;
   // first term -3389/12. replaced! 
   // (taken from dizet6_42.f line 5126)
   m_CA42  = 3533/12.0 - 162.*GMath::Zeta2() - 220.0*GMath::Zeta3() + 
      (-41/6. + 4.*GMath::Zeta2() + 16/3.0*GMath::Zeta3())*m_nf1;

   m_CAL42 = 77/2.0 - 7/3.0*m_nf1;
   
}

void GSM::RadiatorFunctions::UpdateLocalFlags( GReference& /* ref */ )
{
   m_isUpToDate_Update = kFALSE;
}

void GSM::RadiatorFunctions::Update()
{
   if (m_isUpToDate_Update) return;

   // now, it is uptodate (I mean... it will be)
   m_isUpToDate_Update = kTRUE;

   Double_t pi    = TMath::Pi();
      
   Double_t MSMc  = GetQMassRun().GetRunningQuarkMass( GTypes::kCharm );
   Double_t MSMb  = GetQMassRun().GetRunningQuarkMass( GTypes::kBottom );

   m_asMZpi       = GetAlphasMZ()/pi;
   m_aQEDMZpi     = (GConstants::alphaQED()/( 1.0 - GetDAlphaQED().DAlphaQEDMZt() ))/pi;

   m_RatioMcMZ    = GMath::IPow( MSMc/p_MZ,2);
   m_RatioMbMZ    = GMath::IPow( MSMb/p_MZ,2);
   m_RatioMtMZ    = GMath::IPow( p_MZ/p_mt,2);
   
   m_RatioMcMt    = GMath::IPow( MSMc/p_mt,2); 
   m_RatioMbMt    = GMath::IPow( MSMb/p_mt,2);
   
   // light quark corrections
   //   Quadratic corrections (mc and mb term)
   m_LightQu2     = ( m_RatioMcMZ + m_RatioMbMZ )*m_CL3*GMath::IPow(m_asMZpi,3);

   //   Quartic corrections (mc and mb term)
   m_LightQu4Mc   = m_RatioMcMZ*m_RatioMcMZ*(m_C42 - TMath::Log(m_RatioMcMZ))*m_asMZpi*m_asMZpi;
   m_LightQu4Mb   = m_RatioMbMZ*m_RatioMbMZ*(m_C42 - TMath::Log(m_RatioMbMZ))*m_asMZpi*m_asMZpi;

   // Power suppressed top mass correction
   m_Ct2   = m_RatioMtMZ*(44/675.0 - 2/135.0*TMath::Log(m_RatioMtMZ)); 

   // Singlet axial corrections
   m_CMt2 = -37/12.0 + TMath::Log(m_RatioMtMZ) + 7/81.0*m_RatioMtMZ + 0.0132*m_RatioMtMZ*m_RatioMtMZ;
   // + 8/3.0 term not in publications => taken from dizet6_42.f line 5149
   m_CMt3 = -5075/216.0 + 8/3.0 + 23/6.0*GMath::Zeta2() + GMath::Zeta3()
      + 67/18.0*TMath::Log(m_RatioMtMZ) + 23/12.0*GMath::IPow(TMath::Log(m_RatioMtMZ),2);
   
   m_CMt4 = m_CMt3/m_CMt2*m_CMt3*(1.0 - p_DeltaAlphasTheoCMt3_Scale);

   SetUpToDate();
}


// --------------------- Radiator-functions for computing Z->qq ---------------------------

// vector part; Ve = Vector
// eq. (3.6.3) from  hep-ph/9908433
Double_t GSM::RadiatorFunctions::GetRVq( GTypes::Particle ParticleType, Double_t Charge )
{   
   Update();
   
   // heavy quark corrections
   //   Quadratic corrections (mq^2 term only for charm and bottom case)
   Double_t VeHeavyQu2 = m_CV1*m_asMZpi + m_CV2*GMath::IPow(m_asMZpi,2); 
   if (!m_useNLOAlphasOnly) VeHeavyQu2 += m_CV3*GMath::IPow(m_asMZpi,3);

   
   // Quartic corrections
   //   particulary bottom corretions (mq^4 term only for charm and bottom case)
   
   // m_CV40 term is missing in publications, taken from dizet6_42.f line 5128
   Double_t VeHeavyQu4 = m_CV40 + m_CV41*m_asMZpi + m_CV42*GMath::IPow(m_asMZpi,2);
   // this term in dizet6_42.f, but it is double counting see AxialQCD4
   // !!! term makes no difference in result !!!
   // VeHeavyQu4 += 12.*GMath::IPow(m_asMZpi,2);
   Double_t Vector4    = m_CVL42*GMath::IPow(m_asMZpi,2);
   
   // mc and mb terms
   Double_t VectorQCD2 = m_LightQu2;
   Double_t VectorQCD4 = m_LightQu4Mc + m_LightQu4Mb;
   
   if( ParticleType == GTypes::kCharm ){
      VectorQCD2 += m_RatioMcMZ * VeHeavyQu2;
      VectorQCD4 += ( GMath::IPow(m_RatioMcMZ,2)*(VeHeavyQu4 + Vector4*TMath::Log(m_RatioMcMZ))
                      + 12.0*GMath::IPow(m_RatioMbMZ*m_asMZpi,2) );
   }
   else if( ParticleType == GTypes::kBottom){
      VectorQCD2 += m_RatioMbMZ * VeHeavyQu2;
      VectorQCD4 += ( GMath::IPow(m_RatioMbMZ,2)*(VeHeavyQu4 + Vector4*TMath::Log(m_RatioMbMZ))
                      + 12.0*GMath::IPow(m_RatioMcMZ*m_asMZpi,2)
                      - GMath::IPow(m_RatioMbMZ,3)
                      *(8.0 + 16/27.0*(155.0 + 6.0*TMath::Log(m_RatioMbMZ))*m_asMZpi) );
   }
  
   Double_t RadV = ( 1.0 + 3/4.*Charge*Charge*m_aQEDMZpi + m_asMZpi - 1/4.0*Charge*Charge*m_aQEDMZpi*m_asMZpi
                     + (m_C02 + m_Ct2)*GMath::IPow(m_asMZpi,2) 
                     + (m_useNLOAlphasOnly ? 0 : ( m_C03*GMath::IPow(m_asMZpi,3) + 
                                                   m_C04*GMath::IPow(m_asMZpi,4) +
                                                   m_C05*GMath::IPow(m_asMZpi,5) ))
                     + VectorQCD2 + VectorQCD4 );
  
   return RadV;
}
  
// axial part
// eq. (3.6.4) from  hep-ph/9908433
Double_t GSM::RadiatorFunctions::GetRAq( GTypes::Particle ParticleType, Double_t Charge, Double_t T3 )
{
   Update();
   
   // heavy quark corrections
   //   Quadratic corrections (mq^2 term only for charm and bottom case)
   Double_t AxHeavyQu2 = ( m_CA0 + m_CA1*m_asMZpi + m_CA2*GMath::IPow(m_asMZpi,2) 
                           + (m_useNLOAlphasOnly ? 0 : m_CA3*GMath::IPow(m_asMZpi,3)) );

   // Quartic corrections
   //   particulary bottom corretions (mq^4 term only for charm and bottom case)
   Double_t AxHeavyQu4 = m_CA40 + m_CA41*m_asMZpi + m_CA42*GMath::IPow(m_asMZpi,2);
   // this term in dizet6_42.f, but it is double counting see AxialQCD4
   // AxHeavyQu4 += - 12.*GMath::IPow(m_asMZpi,2);
   // !!! term makes no difference in result !!!
   Double_t Axial4     = m_CAL42*GMath::IPow(m_asMZpi,2);
  
   Double_t AxialQCD2  = m_LightQu2;
   Double_t AxialQCD4  = m_LightQu4Mc + m_LightQu4Mb;
   // axial single contribution
   Double_t AxSingle   = 0;
   
   // in publication: AxSingle = 6*m_RatioMbMZ*(3.0 + TMath::Log(m_RatioMtMZ)) ...
   // in publication: AxSingle = ... (8/81. + 1/54.) ...
   // differences corrected with dizet6_42.f
   if( ParticleType == GTypes::kCharm ){
      AxialQCD2 += m_RatioMcMZ * AxHeavyQu2;
      AxialQCD4 += ( GMath::IPow(m_RatioMcMZ,2)*(AxHeavyQu4 + Axial4*TMath::Log(m_RatioMcMZ))
                     - 12.0*GMath::IPow(m_RatioMbMZ*m_asMZpi,2) );
      AxSingle   = ( -6.0*m_RatioMcMZ*(-3.0 + TMath::Log(m_RatioMtMZ))*m_asMZpi*m_asMZpi
                     -10.0*m_RatioMcMt*(8/81.0 - 1/54.0*TMath::Log(m_RatioMtMZ))*m_asMZpi*m_asMZpi ); 
   }
   else if( ParticleType == GTypes::kBottom ){
      AxialQCD2 += m_RatioMbMZ * AxHeavyQu2;
      AxialQCD4 += ( GMath::IPow(m_RatioMbMZ,2)*(AxHeavyQu4 + Axial4*TMath::Log(m_RatioMbMZ))
                     - 12.0*GMath::IPow(m_RatioMcMZ*m_asMZpi,2) );
      AxSingle   = ( -6.0*m_RatioMbMZ*(-3.0 + TMath::Log(m_RatioMtMZ))*m_asMZpi*m_asMZpi
                     -10.0*m_RatioMbMt*(8/81.0 - 1/54.0*TMath::Log(m_RatioMtMZ))*m_asMZpi*m_asMZpi ); 
   }
   
   Double_t RadA = ( 1.0 + 3/4.*Charge*Charge*m_aQEDMZpi + m_asMZpi - 1/4.0*Charge*Charge*m_aQEDMZpi*m_asMZpi
                     + (m_C02 + m_Ct2 - 2.0*T3*m_CMt2)*GMath::IPow(m_asMZpi,2) 
                     + (m_useNLOAlphasOnly ? 0 : ( (m_C03 - 2.0*T3*m_CMt3)*GMath::IPow(m_asMZpi,3) + 
                                                   (m_C04 - 2.0*T3*m_CMt4)*GMath::IPow(m_asMZpi,4) +
                                                   (m_C05)*GMath::IPow(m_asMZpi,5) ))
                     + AxialQCD2 + AxialQCD4 + AxSingle);
   
   return RadA;
}
