/**********************************************************************************
 * Package: GSM                                                                   *
 * Class  : Vertex                                                                *
 *                                                                                *
 * Description:                                                                   *
 *      Auxiliary Theory for Vertex corrections                                   *
 *                                                                                *
 * Sources                                                                        *
 *      - Bardin et al, hep-ph/9709229v1,                                         *
 *      - Bardin et al, CPC. 133 (2001) 229, hep-ph/9908433                       *
 *      - Bardin et al., ZFitter package dizet6_42.f                              *
 *                                                                                *
 * This class also contains code lines ported to C++ from the Fortran package     *
 * ZFITTER                                                                        *
 *                                                                                *
 **********************************************************************************/
#include "TMath.h"

#include "Gfitter/GMath.h"
#include "Gfitter/GConstants.h"
#include "Gfitter/GTheory.h"
#include "Gfitter/GTheoryRef.h"
#include "Gfitter/GParameterRef.h"
#include "Gfitter/GReference.h"

#include "GSM/ZMath.h"
#include "GSM/Vertex.h"

using namespace Gfitter;
using std::complex;

ClassImp(GSM::Vertex)

   GSM::Vertex::Vertex()
   : Gfitter::GAuxTheory(),
     m_isUpToDate_Update( kFALSE )     
{
   SetTheoryName( GetName() );
   SetExistDerivative( kFALSE );

   BookParameter( "MZ"      , & p_MZ );
   BookParameter( "mt"      , & p_mt );
   BookTheory   ( "GSM::WMass" , & t_MW );
}

void GSM::Vertex::UpdateLocalFlags( GReference& /* ref */ )
{
   m_isUpToDate_Update = kFALSE;
}

void GSM::Vertex::Update()
{
   if (m_isUpToDate_Update) return;

   // now, it is uptodate (I mean... it will be)
   m_isUpToDate_Update = kTRUE;

   Double_t MW      = GetMW();
   Double_t MZ      = p_MZ;
   Double_t mt      = p_mt;
   
   Double_t mt2     = mt*mt;
   Double_t MW2     = MW*MW;
   Double_t MZ2     = MZ*MZ;
   Double_t R       = MW2/MZ2;
   
   Double_t Sqr     = TMath::Sqrt(4.0*R - 1.0);
   Double_t Atan    = TMath::ATan(Sqr/(2.0*R - 1.0));
   complex<Double_t>  I (0,1.0);
    
   // Vertex functions in the limit of vanishing fermion masses
   // for Z decay
   // eq. hep-ph/9709229v1 (265)
   // Spence( 1 ) = Li2( 1 ) = Zeta2 
   // see dizet6_42.f line 3311 & 3141
   m_V1ZZ           = - 5.5 - 8.0*( GMath::Zeta2() - TMath::DiLog(2.0) ) + I*imag(V1b(-MZ2, MZ2));
  
   // eq. hep-ph/9709229v1 (265)
   // see also in hep-ph/9908433 (A.4.33)
    // see dizet6_42.f line 3313 & 3142
   m_V1ZW           = ( -3.5 - 2.0*R - (3.0 + 2.0*R)*TMath::Log(R) - 2.0*(1.0 + R)*(1.0 + R)
                        *(GMath::Zeta2() - TMath::DiLog(1.0 + 1.0/R)) + I*imag(V1b(-MZ2,MW2)) );
  
   // eq. hep-ph/9709229v1 (266)
   // see also in hep-ph/9908433 (A.4.34)
    // see dizet6_42.f line 3314
   m_V2ZWW          = ( 2.0/(9.0*R*R) + 43.0/(18.0*R) - 1.0/6.0 - 2.0*R
                        + (-1.0/(12.0*R*R) - 1.5/R + 7.0/3.0 + 2.0*R)*Sqr*Atan - 2.0*R*(2.0 + R)*Atan*Atan );  
  
   Double_t lam = GMath::IPow(MZ2,2) - 4.0*MW2*MZ2;
   
   // -------- for W decay ----------
   // see dizet6_42.f line 3333
   m_V1WZ  = ( -5.0 - 2.0/R + (3.0 + 2.0/R)*TMath::Log(R)
               -2.0*GMath::IPow((1.0+R)/R,2)*(GMath::Zeta2() - TMath::DiLog(1.0 + R)) ); 
   
   // see dizet6_42.f line 3335
   m_V2WWZ = ( -9/(4.0*R) -1/(12.0*R*R) + 23/18.0 
               + (0.5/R - 0.75/(R*R) - 1/(24.0*GMath::IPow(R,3)) + 1.0)*TMath::Log(R)
               - real(ZMath::L(-MW2,MW2,MZ2))*(5/(6.0*R) + 1/(24.0*R*R) + 0.5)/MW2
               + (0.5 + 1.0/R)*lam*GMath::IPow( real(ZMath::J(-MW2,MW2,MZ2)), 2 )
               - (0.5 + 1.0/R)*GMath::IPow( TMath::Log(R), 2 ) );

   // top quark additions
   Double_t RtW      = mt2/MW2;
   Double_t RtW1     = RtW - 1.0;
   
   Double_t J0W = 0, S3W = 0, S30W = 0;
   Double_t J0t = 0, S3t = 0, S30t = 0;

   // s = -MZ2
   ZMath::S3Wana( mt2, MW2, -MZ2, J0W, S3W, S30W);
   ZMath::S3Wana( MW2, mt2, -MZ2, J0t, S3t, S30t);
  
   // Information to this point in hep-ph/9709229 page 88-89,
   // hep-ph/9908433 page 175-177 
   // and dizet6_42.f (ZFitter) line 3374-3393
   Double_t WWv2  = ( - 2.0*R*(2.00 + R)*MZ2*(S3W - S30W) + RtW
                      *((3.0*R*R + 2.5*R - 2.0 - (2.0*R - 0.5)*RtW +R*(0.5 - R)*RtW*RtW)*MZ2*S3W
                        -(R + 1.00 - (0.5 - R)*RtW)*(J0W - 2.0) +(2.0*R + 3/(2.0*RtW1*RtW1) 
                                                                  - 2.0/RtW1 + 0.5 - (0.5 - R)*RtW)*TMath::Log(RtW)
                        -(R + 3/(2.0*RtW1) + 0.75 - (0.5 - R)*RtW) + 0.25/R*(J0W - 3.0)*1.0) );
   Double_t WWv11 = ( + 2.0*(1.0 + R)*(1.0 + R)*MZ2*(S3t-S30t) + (2.0*R + 3.0)*(J0t + TMath::Log(RtW) + TMath::Log(R))
                      - RtW*( R*(3.0*R + 2.0 - RtW - R*RtW*RtW)*MZ2*S3t + (R + 0.5 + R*RtW)*(J0t + TMath::Log(RtW) - 2.0)
                              - (2.00*R + 3/(2.0*RtW1*RtW1) - 2.0/RtW1 + 0.5 + R*RtW)*TMath::Log(RtW)
                              + R + 3/(2.0*RtW1) + 5/4.0 + R*RtW ) ); 
   Double_t WWv12 = ( -RtW*(R*(2.0 + R - 2.0*R*RtW + R*RtW*RtW)*MZ2*S3t
                            -(0.5 - R + R*RtW)*(J0t + TMath::Log(RtW) - 1.0) + R*RtW*TMath::Log(RtW) ) );
  
   // see dizet6_42.f line 3322
   Double_t v_tb  = 1;
   m_Vtb          = v_tb*v_tb*( R*WWv2 - 0.5*(1.0 - 2.0*(1.0 - R)*(2/3.0))*WWv11 - 0.5*WWv12 );
                                                   
   // now parameters are up-to-date
   SetUpToDate();
}
 

// hep-ph/9709229v1 (265) - (267)
// deal with the imaginary part
   // see dizet6_42.f line 918-932
complex<Double_t> GSM::Vertex::V1b( Double_t Q2, Double_t MV2 )
{   
   Double_t r      = Q2/MV2;
   Double_t Logr   = TMath::Log(TMath::Abs(r));
   Double_t Rev1b  = ( -3.5 + 2.0/r + (3.0-2.0/r)*Logr + 2.0 * GMath::IPow(1.0-1.0/r,2) * 
                       (TMath::DiLog(1.0-r) - GMath::Zeta2()) );
   Double_t Aiv1b = 0;

   if (Q2 < 0) Aiv1b = TMath::Pi()*(-3.0+2.0/r+2.0*GMath::IPow((1.0-1.0/r),2)*TMath::Log(1.0-r));
   complex<Double_t> v1b(Rev1b,Aiv1b);
    
   return v1b;
}




