/**********************************************************************************
 * Package: GSM                                                                   *
 * Class  : WZFitter                                                              *
 *                                                                                *
 * Description:                                                                   *
 *      Auxiliary Theory for electroweak formfactor for W decay                   *
 *      W Boson decay / total width                                               *
 *                                                                                *
 * Sources:                                                                       *
 *      - Bardin, Degrassi, The Standard Model in the Making, Oxford 1999         *
 *      - Bardin et al., ZFitter package dizet6_42.f                              *
 *                                                                                *
 * This class also contains code lines ported to C++ from the Fortran package     *
 * ZFITTER                                                                        *
 *                                                                                *
 **********************************************************************************/
#include "TMath.h"

#include "Gfitter/GMath.h"
#include "Gfitter/GConstants.h"
#include "Gfitter/GTheory.h"
#include "Gfitter/GTheoryRef.h"
#include "Gfitter/GParameterRef.h"
#include "Gfitter/GReference.h"

#include "GSM/WZFitter.h"
#include "GSM/FermionPart.h"
#include "GSM/BosonPart.h"
#include "GSM/Vertex.h"

using namespace Gfitter;
using std::complex;

ClassImp(GSM::WZFitter)

GSM::WZFitter::WZFitter()
   : WBase(), 
     m_isUpToDate_Update( kFALSE )     
{
   SetTheoryName( GetName() );
   SetExistDerivative( kFALSE );

   BookTheory   ( "GSM::FermionPart"             , &t_fermionPart );
   BookTheory   ( "GSM::BosonPart"               , &t_bosonPart );
   BookTheory   ( "GSM::Vertex"                  , &t_Vertex );
   BookTheory   ( "GSM::RadiatorFunctions"       , &t_RadFun );
   BookTheory   ( "GSM::WMass"                      , & t_MW );
}

void GSM::WZFitter::UpdateLocalFlags( GReference& /* ref */ )
{
   m_isUpToDate_Update = kFALSE;
}

void GSM::WZFitter::Update()
{
   if (m_isUpToDate_Update) return;

   // now, it is uptodate (I mean... it will be)
   m_isUpToDate_Update = kTRUE;

   Double_t MW  = GetMW();
   Double_t MZ  = p_MZ;
   Double_t MW2 = MW*MW;
   Double_t MZ2 = MZ*MZ;
        
   m_R          = MW2/MZ2;
   
   // bosonic + ferminoic contribution
   // see dizet6_42.f line 3408-3410
   m_W_MW  = real( GetBosonPart().GetWbAtMW()  + GetFermionPart().GetWfAtMW() );
   m_W0    = real( GetBosonPart().GetWbAt0()   + GetFermionPart().GetWfAt0() );
   m_WF_MW = real( GetBosonPart().GetWbFAtMW() + GetFermionPart().GetWfFAtMW() );
   
   // eq. (10.67) of The Standard Model in the Making
   // without mass-corrections, 
   // they are neglibile in relation to the W mass
    // see dizet6_42.f line 2849
   Double_t Gamma0 = ( p_GF*Gfitter::GMath::IPow( GetMW(),3 )
                       / (6.0*TMath::Pi()*Gfitter::GMath::Sqrt2()) );
  
   // decay into leptons
   Double_t RhoLep = GetRhoW( 1 );
  
   // decay into hadrons
   Double_t RhoQua = GetRhoW( Gfitter::GMath::TwoThird() );
   // some QCD corrections
   Double_t RadVu  = GetRadFun().GetRVq( Gfitter::GTypes::kUp, Gfitter::GMath::TwoThird() );

   // see dizet6_42.f line 2860-2869
   m_gammaW = Gamma0*( RhoLep*3.0 + RhoQua*6.0*RadVu );
  
   // now, parameters are up-to-date
   SetUpToDate();
}

// rho for W decay, just 1-loop, because it is impossible 
// to define QED-gauge invariant subset of diagrams
Double_t GSM::WZFitter::GetRhoW( Double_t Charge )
{
   Update();
   Double_t ChUpDo = Charge*(1.0 - Charge);
   
   // for better understanding
   // eq. (10.71) of The Standard Model in the Making
   // one and two point functions are replaced
   // by fermionic and bosonic contribution
   
   // see dizet6_42.f line 3411
   Double_t rho    = ( 1.0 + GConstants::alphaQED()/(4.0*TMath::Pi()*(1.0-m_R))
                       *(m_W_MW - m_W0 + m_WF_MW - 7.0 + 5/8.0*m_R*(1.0+m_R) 
                         - 9/4.0*m_R/(1.0-m_R)*TMath::Log(m_R) 
                         + 0.75/m_R + 3.0*m_R - 3.0/m_R*(1.0-m_R)*(1.0-m_R)*ChUpDo
                         + (0.5/m_R - 1.0 - 2.0*(1.0-m_R)*(1.0-m_R)/m_R*ChUpDo)*GetVertex().GetV1WZ()
                         + 2.0*m_R*GetVertex().GetV2WWZ() + 2.0*(1.0-m_R)
                         *(77/12.0 - 2/3.0*GMath::IPow( TMath::Pi(),2 ) + 109/36.0 - 3/2.0*ChUpDo)) );
   return rho;
}

