/**********************************************************************************
 * Package: GSM                                                                   *
 * Class  : Z0ZFitter                                                             *
 *                                                                                *
 * Description:                                                                   *
 *      Auxiliary Theory of the ZFitter option using the OMS approach             *
 *      Computes effective weak mixing angle and Partial Z widths                 *
 *                                                                                *
 * Sources:                                                                       *
 *      - Bardin et al, hep-ph/9709229v1                                          *
 *      - Bardin et al, hep-ph/9412201v3                                          *
 *      - Bardin et al,,Comput.Phys.Commun. 133 (2001) 229, hep-ph/9908433        *
 *      - Bardin et al., ZFitter package dizet6_42.f                              *
 *                                                                                *
 * This class also contains code lines ported to C++ from the Fortran package     *
 * ZFITTER                                                                        *
 *                                                                                *
 **********************************************************************************/
#include "TMath.h"

#include "Gfitter/GMath.h"
#include "Gfitter/GConstants.h"
#include "Gfitter/GTheory.h"
#include "Gfitter/GTheoryRef.h"
#include "Gfitter/GParameterRef.h"
#include "Gfitter/GReference.h"
#include "Gfitter/GVariable.h"
#include "Gfitter/GStore.h"

#include "GSM/Z0ZFitter.h"
#include "GSM/ZMath.h"

#include "GSM/MW.h"

using namespace Gfitter;
using std::complex;
using namespace Gfitter::GTypes;

ClassImp(GSM::Z0ZFitter)

GSM::Z0ZFitter::Z0ZFitter()
   : Z0Base(),
     m_isUpToDate_Update( kFALSE )
{
   SetTheoryName( GetName() );
   SetExistDerivative( kFALSE );

   const TString& omstype = gStore()->GetVariable( "GSMFlags::OMSType" )->GetStringValue();
   m_logger << kINFO << "Using OMS type: \"" << omstype << "\"" << GEndl;
   
   if      (omstype == "OMS1" )  m_OMSType  = OMS1;
   else if (omstype == "OMS1_2") m_OMSType  = OMS1_2;
   else if (omstype == "OMS2")   m_OMSType  = OMS2;
   else {
      m_logger << kFATAL << "unknown value for \"GSMFlags::OMStype\": \"" << omstype << "\""
               << ". Possible are: \"OMS1\", \"OMS1-2\" \"OMS2\""
               << GEndl;
   }

   const TString& logMH = gStore()->GetVariable( "GSMFlags::logMH" )->GetStringValue();
   m_logger << kINFO << "Using logMH: \"" << logMH << "\"" << GEndl;
   
   if      (logMH == "Yes" ) m_logMH  = kTRUE;
   else if (logMH == "No" )  m_logMH  = kFALSE;
   else {
      m_logger << kFATAL << "unknown value for \"GSMFlags::logMH\": \"" << logMH << "\""
               << ". Possible are: \"Yes\" and \"No\"\""
               << GEndl;
   }

   BookParameter( "DeltaOMS"                   , &p_DeltaOMS );  

   BookTheory   ( "GSM::Vertex"                  , &t_Vertex );
   BookTheory   ( "GSM::QCDCorrections"          , &t_QCDCorr );
   BookTheory   ( "GSM::FermionPart"             , &t_FermionPart );
   BookTheory   ( "GSM::BosonPart"               , &t_BosonPart );
   BookTheory   ( "GSM::EW2Loop"                 , &t_EW2Loop );
   BookTheory   ( "GSM::RadiatorFunctions"       , &t_radFun );

   BookTheory   ( "GSM::RunningAlphaQCD"       , &t_AlphasRun );
   BookTheory   ( "GSM::WMass"                    , &t_MW );
}

void GSM::Z0ZFitter::Initialise()
{
   //non-factorizable EW*QCD corrections
   m_EWQCD[0] = -0.113e-3;
   m_EWQCD[1] = -0.160e-3;
   m_EWQCD[2] =  m_EWQCD[0];
   m_EWQCD[3] =  m_EWQCD[1];
   m_EWQCD[4] = -0.040e-3;
}

void GSM::Z0ZFitter::UpdateLocalFlags( GReference& /* ref */ )
{
   m_isUpToDate_Update = kFALSE;
}

void GSM::Z0ZFitter::Update()
{
   if (m_isUpToDate_Update) return;

   // now, it is uptodate (I mean... it will be)
   m_isUpToDate_Update = kTRUE;

   Double_t pi        = TMath::Pi();

   Double_t MW        = GetMW();
   Double_t MZ        = p_MZ;
   Double_t mt        = p_mt;
   Double_t MH        = GetMH().GetValue(); 
   if (m_logMH) MH = TMath::Exp( GetMH().GetValue() );

   // alphas at MZ and mtop
   Double_t alphasMZ  = GetAlphasRun().EvolveAlphas(p_MZ);;

   if (TMath::IsNaN(p_mt)) m_logger << kFATAL << "<Z0ZFitter::Update> p_mt is NaN !" << GEndl;
   Double_t alphasTop = GetAlphasRun().EvolveAlphas(p_mt);;
   
   m_R = MW*MW/(MZ*MZ);
 
   // leading order contribution
   // W and Z self-energies at different scales
   // MPhoZAtMZ = photon Z mixing function
   // bosonic + ferminoic part
   // see dizet6_42.f line 3133-3137
   m_WAtMW       =   GetBosonPart().GetWbAtMW()     + GetFermionPart().GetWfAtMW();
   m_ZFAtMZ      =   GetBosonPart().GetZbFAtMZ()    + GetFermionPart().GetZfFAtMZ();
   m_WAt0        =   GetBosonPart().GetWbAt0()      + GetFermionPart().GetWfAt0();
   m_MPhoZAtMZ   =   GetBosonPart().GetMbPhoZAtMZ() + GetFermionPart().GetMfPhoZAtMZ();
   m_ZAtMZ       = ( GetBosonPart().GetZbAtMZ()     + GetFermionPart().GetZfAtMZ() - 1/(4.0*pi)
                     * GConstants::alphaQED()*pow( GetFermionPart().GetMfPhoZAtMZ(), 2 ) );
  
   // QCD corrections 
   // see dizet6_42 line 1858 & 1859
   Double_t v2u = GMath::IPow(1.0-4.0*2/3.0*(1.0-m_R),2);
   Double_t v2d = GMath::IPow(1.0-4.0*1/3.0*(1.0-m_R),2);
 
   // light and heavy quarks
   // (3.415) of hep-ph/9908433
   // The Standard Model in the Making page 413-414
   // see dizet6_42.f line 3181
   m_rhoQCD     = ( GConstants::alphaQED()/(1.0*pi)
                    * alphasTop/pi*real( GetQCDCorr().GetRMQCD() ) 
                    + GConstants::alphaQED()/(1.0*pi)
                    * alphasMZ/(8.0*pi*m_R*(1.0-m_R))
                    *(v2u+v2d+2.0) );
  
   // (3.416) of hep-ph/9908433
   // The Standard Model in the Making page 413-414
   // see dizet6_42.f line 3183
   m_kapQCD     = ( GConstants::alphaQED()/(1.0*pi)
                    * alphasTop/pi*real( GetQCDCorr().GetKMQCD() ) 
                    + GConstants::alphaQED()/(1.0*pi)
                    * alphasMZ/pi
                    * m_R/(2.0*(1.0-m_R)*(1.0-m_R))*TMath::Log(m_R) );
   
     
   // additional corrections (reminders)
   // see dizet6_42.f line 3050
   Double_t WZCorr  = m_R/(1.0-m_R)*real( m_WAtMW - m_ZAtMZ )/(1.0 -m_R);
   // see dizet6_42.f line 3052
   Double_t Scale   = GConstants::alphaQED()/(4*pi*(1.0-m_R))*(41/6.0 - 11/3.0*m_R)*TMath::Log(m_R);
   // see dizet6_42.f line 3055
   Double_t TopCon  = p_GF*mt*mt/( 8.0*GMath::Sqrt2()*pi*pi );
   // see dizet6_42.f line 3055
   Double_t RhoTop4 = 3.0*TopCon*(1.0 + TopCon*ZMath::BarbMc( MH/mt ));
   // see dizet6_42.f line 3064
   Double_t TBCor0  = 3.0*TopCon*ZMath::Afmt3( alphasTop, mt*mt, MZ*MZ, (1.0-m_R) );
   // see dizet6_42.f line 3226
   Double_t DkTBCor = 3.0*TopCon*ZMath::TbQCD( alphasTop, mt*mt, MZ*MZ, (1.0-m_R) );

   // see dizet6_42.f line 3051
   m_Renorm         = GMath::Sqrt2()*p_GF*MZ*MZ*(1.0-m_R)*m_R/(pi*GConstants::alphaQED());
   // see dizet6_42.f line 3053
   m_KapCorr        = ( GConstants::alphaQED()/(4*pi)*WZCorr + Scale 
                        + 0.75*GConstants::alphaQED()/(4*pi*(1.0-m_R)*(1.0-m_R))*mt*mt/(MZ*MZ) );
  
   // see dizet6_42.f line 3054
   m_RhoTop         = 0.75*GConstants::alphaQED()/(4*pi*m_R*(1.0-m_R))*mt*mt/(MZ*MZ);
   // see dizet6_42.f line 3067
   m_TBCorL         = -GConstants::alphaQED()/(4*pi)*alphasTop/pi*mt*mt/(MW*MW*(1.0-m_R))*(0.5+pi*pi/6.0);
       
   // see dizet6_42.f line 3071-3072
   m_RhoRemAdd      = -1.0 - m_TBCorL +  (1.0-m_R)/m_R*m_KapCorr - m_RhoTop;
   m_KapRemAdd      = -1.0 - m_R/(1.0-m_R)*(m_TBCorL + m_RhoTop) + m_KapCorr + DkTBCor;

   // see hep-ph/9908433 page 77, DROBAR
   // see dizet6_42.f line 2300
   m_DRhoBar        = ( TBCor0 - (GConstants::alphaQED()/(4*pi)*WZCorr + Scale)
                         *GMath::Sqrt2()*p_GF*MZ*MZ*GMath::IPow( 1-m_R , 2 )/pi*1/GConstants::alphaQED()
                         + GetEW2Loop().GetDrho2L() );
   // see hep-ph/9908433 page 77, DROBLO
   // see dizet6_42.f line 2302  
   m_DRhoBarLO      = (  - (GConstants::alphaQED()/(4*pi)*WZCorr + Scale)
                          *GMath::Sqrt2()*p_GF*MZ*MZ*GMath::IPow( 1-m_R , 2 )/pi*1/GConstants::alphaQED() );
 
  
   // eq. (2.4.23) of hep-ph/9908433
   // bosonic reminder term
   // see dizet6_42.f line 2176-2178
   m_DrRem          =  ( GConstants::alphaQED()/(4*pi)
                         *(-2.0/3.0 + 1.0/(1.0-m_R)
                           *(real( GetBosonPart().GetWbAt0() ) - real( GetBosonPart().GetWbAtMW() )
                             - 5.0/8.0*m_R*(1.0+m_R) + 11.0/2.0 + 9.0/4.0*m_R/(1.0-m_R)*TMath::Log(m_R)))
                         - GConstants::alphaQED()/(4*pi*(1.0-m_R))*(1.0/6.0 + 7.0*m_R)*TMath::Log(m_R) );
   // fermionic reminder term
   m_DrRem         += ( GConstants::alphaQED()/(4*pi*(1.0-m_R))
                        *( real( GetFermionPart().GetWfAt0() ) - real( GetFermionPart().GetWfAtMW() ) )
                        + GConstants::alphaQED()/(pi*(1.0-m_R))*TMath::Log(m_R) );
  
   // corrections for Z->bb
   // there isn't a full two loop calculation
   // see dizet6_42.f line 3789-3804
   Double_t xt      = GConstants::GF()*mt*mt/(8*GMath::Sqrt2()*pi*pi);
   m_taub1          = -2.0*xt*( 1.0 - pi/3.0*alphasTop );
   m_taub2          = -2.0*xt*( xt*ZMath::BarbMb( MH/mt ) );
 
   // see dizet6_42.f line 3069-3070
   m_RhoCL          = 1/( 1.0 - RhoTop4 - TBCor0 );
   m_KapCL          = ( 1.0 + m_R/(1.0-m_R)*(RhoTop4 + TBCor0) - m_KapCorr*m_Renorm );
   
   // see dizet6_42.f line 2972
   m_CorrBB         = GConstants::alphaQED()/(8.0*pi*(1.0-m_R))*mt*mt/(MW*MW);
  
   // for Z decay
   // see dizet6_42.f line 3167
   m_kapmix         = ( 1/pi*GConstants::alphaQED()
                        *alphasMZ/(24.0*(1-m_R)*(1-m_R))*(2.0*(1-m_R)-1.0) );
 

   SetUpToDate();
}

// effective weak mixing angle
// see also class Sin2ThetaF.h
Double_t GSM::Z0ZFitter::GetSin2Eff( GTypes::Particle ParticleType )
{   
   Double_t sineff = 0;

   sineff = GetSinEffF().GetSin2ThetaF( ParticleType );
   
   return sineff;
}

// eq. (2.4.21) of hep-ph/9908433
// formfactor rho_Z_f
Double_t GSM::Z0ZFitter::rhof( GTypes::Particle ParticleType, Double_t Charge )
{
   Update();
   Double_t rho    = 0;
   Double_t RhoRem = 0;

   switch(ParticleType){
   case kElectron:
   case kMuon:
   case kTau:
   case kNeutrino:
   case kUp:
   case kDown:
   case kStrange:
   case kCharm:
      // see dizet6_42.f line 3153 & 3190 & 3231
      RhoRem = ( sqrt(GMath::IPow(1.0 + real(rhofLO(ParticleType,Charge)),2) 
                      + GMath::IPow(imag(rhofLO(ParticleType,Charge)),2)) + m_rhoQCD + m_RhoRemAdd );
     
      // see dizet6_42.f line 3248-3258
      switch(m_OMSType){
      case OMS1:
         rho = ( (1.0 + m_Renorm*RhoRem + GetEW2Loop().GetRhoRem( TMath::Abs(Charge)))
                 /( 1.0 - m_DRhoBar*(1.0 - m_Renorm*m_DrRem) ) );
         break;
      case OMS1_2:
         rho = ( (1.0 + m_Renorm*RhoRem)/(1.0 - m_DRhoBar*(1.0 - m_Renorm*m_DrRem))
                 + GetEW2Loop().GetRhoRem( TMath::Abs(Charge) ) );
         break;
      case OMS2:
         rho = ( 1.0 + m_DRhoBar - m_DRhoBarLO*m_Renorm*m_DrRem + m_DRhoBarLO*m_DRhoBarLO
                 +m_Renorm*RhoRem*(1.0 + m_DRhoBarLO) + GetEW2Loop().GetRhoRem( TMath::Abs(Charge) ) );
         break;
      default:
         m_logger << kFATAL << "unknown value for \"m_QMSType\": \"" << GetOMSType() << "\""
                  << GEndl;
      }
      break;
   case kBottom:
          // bottoms don't have full two loop corrections
          // and taub is handling in 
          // see dizet6_42.f line 2989 & 3071
      RhoRem = real(rhofLO( ParticleType, Charge )) + m_rhoQCD + 2.0*m_CorrBB - m_RhoTop - m_TBCorL;
      
      // see dizet6_42.f line 3082-3101
      switch(m_OMSType){
      case OMS1:
         rho = 1.0/(1.0/(m_RhoCL*GMath::IPow( (1 + m_taub1 + m_taub2),2 )) - 1.0*RhoRem);
         break;
      case OMS1_2:
         rho = ( m_RhoCL*GMath::IPow( (1 + m_taub1 + m_taub2),2 )
                 *(1.0 + RhoRem*m_RhoCL*GMath::IPow((1 + m_taub1 + m_taub2),2)) ) ;
         break;
      case OMS2:
         rho = m_RhoCL*GMath::IPow( (1 + m_taub1 + m_taub2),2 )*(1.0 + RhoRem);
         break;
      default:
         m_logger << kFATAL << "unknown value for \"m_QMSType\": \"" << GetOMSType() << "\""
                  << GEndl;
      }
      break;
   default:
      m_logger << kFATAL << "unknown value for \"ParticleType\": \"" << GTypes::GetName(ParticleType) << "\""
               << GEndl;
   }
 
   return rho + rho*p_DeltaOMS;
}

// eq. (2.4.24) of hep-ph/9908433
// formfactor kappa_Z_f
Double_t GSM::Z0ZFitter::kappaf( GTypes::Particle ParticleType, Double_t Charge )
{
   Update();
   Double_t kappa = 0;
   Double_t KapRem = 0;
   Double_t ka = 0,Da = 0;

   switch(ParticleType){
   case kElectron:
   case kMuon:
   case kTau:
   case kNeutrino:
   case kUp:
   case kDown:
   case kStrange:
   case kCharm:
      // see dizet6_42.f line 3154 & 3191 & 3232
      KapRem   = ( 1.0 + real(kappafLO(ParticleType,Charge)) + m_kapQCD + m_KapRemAdd ); 
      // see dizet6_42.f line 3261-3271
      switch(m_OMSType){
      case OMS1:
         kappa = ( (1.0 + m_Renorm*KapRem + GetEW2Loop().GetDkdRem( TMath::Abs( Charge ) ))
                   *(1.0 + m_R/(1.0-m_R)*m_DRhoBar*(1.0 - m_Renorm*m_DrRem)) );
         break;
      case OMS1_2:
         kappa = ( (1.0 + m_Renorm*KapRem)*(1.0 + m_R/(1.0-m_R)*m_DRhoBar*(1.0 - m_Renorm*m_DrRem))
                   + GetEW2Loop().GetDkdRem( TMath::Abs( Charge ) ) );
         break;
      case OMS2:
         kappa = ( 1.0 + m_R/(1.0-m_R)*m_DRhoBar - m_R/(1.0-m_R)*m_DRhoBarLO*m_Renorm*m_DrRem
                   + m_Renorm*KapRem*(1.0 + m_R/(1.0-m_R)*m_DRhoBarLO) 
                   + GetEW2Loop().GetDkdRem( TMath::Abs( Charge ) ) );
         break;
      default:
         m_logger << kFATAL << "unknown value for \"m_QMSType\": \"" << GetOMSType() << "\""
                  << GEndl;
      }
      Da      = GetDAlphaQED().DAlphaQEDMZ()*GConstants::alphaQED()/(4.0*TMath::Pi());
      ka      = GMath::IPow( GConstants::alphaQED()/(1.0 - Da), 2 )*35/18.0*(1.0 - 8/3.0*kappa*(1.0-m_R));
      kappa  += ka/(1.0-m_R); 
      break;
   case kBottom:
      // bottoms don't have full two loop corrections 
      // and taub is handling in 
      // see dizet6_42.f line 2990 & 3072
      KapRem   = ( real(kappafLO( ParticleType, Charge )) + m_kapQCD - m_CorrBB 
                   - m_R/(1.0-m_R)*(m_RhoTop + m_TBCorL) + m_KapCorr );
       
       // see dizet6_42.f line 3082-3101
      switch(m_OMSType){
      case OMS1:
         kappa = m_KapCL/(1 + m_taub1 + m_taub2)*( 1.0 + 1.0*KapRem );
         break;
      case OMS1_2:
         kappa = m_KapCL/(1 + m_taub1 + m_taub2) + KapRem;
         break;
      case OMS2:
         kappa = m_KapCL/(1 + m_taub1 + m_taub2) + KapRem;
         break;
      default:
         m_logger << kFATAL << "unknown value for \"m_QMSType\": \"" << GetOMSType() << "\""
                  << GEndl;
      }
      break;
   default:
      m_logger << kFATAL << "unknown value for \"ParticleType\": \"" << GTypes::GetName(ParticleType) << "\""
               << GEndl;
   }
 
   return kappa + p_DeltaOMS*kappa;
}

// leading order of rhof defined in (253) of hep-ph/9709229v1
complex<Double_t> GSM::Z0ZFitter::rhofLO( GTypes::Particle ParticleType, Double_t Charge )
{ 
   complex<Double_t> rho = 0;
   
   switch(ParticleType){
   case kElectron:
   case kMuon:
   case kTau:
   case kNeutrino:
   case kUp:
   case kDown:
   case kStrange:
   case kCharm:
      // see dizet6_42.f line 3147
      rho = ( GConstants::alphaQED()/(4*TMath::Pi()*(1.0-m_R))*
              ( real(m_ZAtMZ + m_ZFAtMZ) - m_WAt0 + 5.0/8.0*m_R*(1.0+m_R) 
                - 11.0/2.0 - 9.0*m_R/(4.0*(1.0-m_R))*TMath::Log(m_R) + 2.0*Uff(ParticleType, Charge)) );
      break;
   case kBottom:
      // see dizet6_42.f line 2907
      rho = ( GConstants::alphaQED()/(4*TMath::Pi()*(1.0-m_R))*
              ( real(m_ZAtMZ) + real(m_ZFAtMZ) - real(m_WAt0) + 5.0/8.0*m_R*(m_R + 1.0) 
                - 11.0/2.0 - 9.0*m_R/(4.0*(1.0-m_R))*TMath::Log(m_R) + Uff(ParticleType, Charge)) );
      break;
   default:
      m_logger << kFATAL << "false input for ParticleType" << GTypes::GetName(ParticleType) << GEndl;
      break;
   }

   return rho;
}

// leading order to kappaf defined in (254) of hep-ph/9709229v1
complex<Double_t> GSM::Z0ZFitter::kappafLO( GTypes::Particle ParticleType, Double_t Charge )
{
   complex<Double_t> kappa = 0;
   
   switch(ParticleType){
   case kElectron:
   case kMuon:
   case kTau:
   case kNeutrino:
   case kUp:
   case kDown:
   case kStrange:
   case kCharm:
      // see dizet6_42.f line 3150
      kappa = ( GConstants::alphaQED()/(4*TMath::Pi()*(1-m_R))*
                ( m_R/(1-m_R)*real(m_ZAtMZ - m_WAtMW) 
                  + m_MPhoZAtMZ + GMath::IPow((1-m_R),2)/m_R*Charge*Charge*GetVertex().GetV1ZZ() 
                  - Uff(ParticleType, Charge) ) ); 
      break;
   case kBottom:
      // see dizet6_42.f line 2910
      kappa = ( GConstants::alphaQED()/(4*TMath::Pi()*(1-m_R))*
                ( m_R/(1-m_R)*( real( m_ZAtMZ ) - real( m_WAtMW ) ) 
                  + real( m_MPhoZAtMZ ) + GMath::IPow((1-m_R),2)/m_R*Charge*Charge*real( GetVertex().GetV1ZZ() ) 
                  - 0.5*Uff(ParticleType, Charge) ) );
      break;
   default:
      m_logger << kFATAL << "false input for ParticleType : " << GTypes::GetName(ParticleType) << GEndl;
      break;  
   }
   
   return kappa;
   
}

// eq. (255) of hep-ph/9709229v1
complex<Double_t> GSM::Z0ZFitter::Uff( GTypes::Particle ParticleType, Double_t Charge )
{   
   complex<Double_t> uff = 0;
  
   switch(ParticleType){
   case kElectron:
   case kMuon:
   case kTau:
   case kNeutrino:
   case kUp:
   case kDown:
   case kStrange:
   case kCharm:
      // see dizet6_42.f line 3144
      uff = ( 0.25/m_R*(1.0 - 6.0*TMath::Abs(Charge)*(1.0-m_R) 
                        + 12.0*Charge*Charge*(1.0-m_R)*(1.0-m_R))*GetVertex().GetV1ZZ()
              + (0.5 - m_R - TMath::Abs(Charge)*(1.0-m_R))*GetVertex().GetV1ZW() 
              + m_R*GetVertex().GetV2ZWW() );
      break;
   case kBottom:
      // see dizet6_42.f line 2903
      uff = ( 0.5/m_R*(1.0 - 6.0*TMath::Abs(Charge)*(1.0-m_R) 
                       + 12.0*Charge*Charge*(1.0-m_R)*(1.0-m_R))*real(GetVertex().GetV1ZZ())
              + (1.0 - 2.0*m_R - 2.0*TMath::Abs(Charge)*(1.0-m_R))*real(GetVertex().GetV1ZW()) 
              + 2.0* m_R*real(GetVertex().GetV2ZWW()) + 2.0*GetVertex().GetVtb() );
      break;
   default:
      m_logger << kFATAL << "false input for ParticleType : " << GTypes::GetName(ParticleType) << GEndl;
      break;
  
   }
   
   return uff;
} 

// partial decay width of the Z Boson
// eq. (2.49) and (2.4.10) of hep-ph/9908433
Double_t GSM::Z0ZFitter::GammaZff( GTypes::Particle ParticleType, Double_t Charge, Double_t mf )
{
   Update();
   
   Double_t Gamma0     = ( p_GF*GMath::IPow(p_MZ,3)/
                           (12.0*TMath::Pi()*GMath::Sqrt2()) );
   Double_t aQEDMZ     = GConstants::alphaQED()/( 1 - GetDAlphaQED().DAlphaQEDMZt() );
   Double_t Rhof       = rhof( ParticleType, Charge );
   Double_t T3         = 0.5*TMath::Sign(1.0,Charge+0.1);
   Double_t SqrRatio   = 1;
   Double_t RatioMfMZ  = 0;
   Double_t RadV       = 0;
   Double_t RadA       = 0;
   Double_t AddEWQCD   = 0;
   Int_t ncf           = 0;
   Int_t nf            = 0;

   switch(ParticleType){
   case kElectron :
   case kMuon     :
   case kTau      :
   case kNeutrino :
      RadV       = 1.0 + 0.75*aQEDMZ/TMath::Pi()*Charge*Charge;
      RadA       = RadV;
      RatioMfMZ  = GMath::IPow(mf/p_MZ,2);
      SqrRatio   = sqrt(1.0-4.0*RatioMfMZ);
      ncf        = 1;
      break;
   case kBottom  :
      ++nf;
   case kStrange :
      ++nf;
   case kCharm   :
      ++nf;
   case kDown    :
      ++nf;
   case kUp      : 
      ++nf;
      RadV       = GetRadFun().GetRVq( ParticleType, Charge);
      RadA       = GetRadFun().GetRAq( ParticleType, Charge, T3);
      RatioMfMZ  = 0;
      SqrRatio   = 1;
      ncf        = 3;
      AddEWQCD   = m_EWQCD[nf-1];
      break;
   default:
      m_logger << kFATAL << "false input for ParticleType : " << GTypes::GetName(ParticleType) << GEndl;
      break;
   } 
  
   // see dizet6_42.f line 3169
   Double_t KappaImag  = imag(kappafLO(ParticleType,Charge)) + m_kapmix;
   if( ParticleType == kBottom ) KappaImag = 0;
   Double_t sineff     = GetSin2Eff(ParticleType);
   Double_t RatioCplgs = 1.0 - 4.0*TMath::Abs(Charge)*sineff;
  
   // see dizet6_42.f line 2734-2793
   Double_t Gamma      = ( Gamma0*Rhof*SqrRatio*
                           ( (1.0+2.0*RatioMfMZ)
                             *( (RatioCplgs*RatioCplgs + 16.0*(1-m_R)*(1-m_R)*
                                 Charge*Charge*KappaImag*KappaImag)*RadV 
                                + RadA)/2.0 - 3.0*RatioMfMZ*RadA) );
   
   return Gamma*ncf + AddEWQCD;

}


