/**********************************************************************************
 * Package: GSM                                                                   *
 * Class  : Z0ZFitter                                                             *
 *                                                                                *
 * Description:                                                                   *
 *      Auxiliary Theory of the ZFitter option using the OMS approach             *
 *      Computes effective weak mixing angle and Partial Z widths                 *
 *                                                                                *
 **********************************************************************************/
#ifndef GSM_Z0ZFitter
#define GSM_Z0ZFitter

#include <complex>

#include "Gfitter/GConstants.h"

#include "GSM/QCDCorrections.h"
#include "GSM/FermionPart.h"
#include "GSM/BosonPart.h"
#include "GSM/Vertex.h"
#include "GSM/EW2Loop.h"

#include "GSM/Z0Base.h"
#include "GSM/RunningAlphaQCD.h"
#include "GSM/RadiatorFunctions.h"
#include "GSM/WMass.h"

using namespace Gfitter;

class Gfitter::GReference;
class Gfitter::GParameterRef;
class Gfitter::GTheoryRef;

namespace GSM {

   class Z0ZFitter : public GSM::Z0Base {

   public:
      
      Z0ZFitter();
      virtual ~Z0ZFitter() {}
      
      void Initialise();

      // accessors
      virtual Double_t GetSin2Eff(Gfitter::GTypes::Particle p);
      virtual Double_t GetGaZelectron();
      virtual Double_t GetGaZmuon();
      virtual Double_t GetGaZtau();
      virtual Double_t GetGaZneutrino();
      virtual Double_t GetGaZup();
      virtual Double_t GetGaZdown();
      virtual Double_t GetGaZstrange();
      virtual Double_t GetGaZcharm();
      virtual Double_t GetGaZbottom();
      virtual Double_t GetDeltaS(){ return 0; };
      virtual Double_t GetDeltaT(){ return 0; };
     
      // effective vector and axial couplings
      virtual Double_t GetgAeff( Gfitter::GTypes::Particle particle, Double_t T3, Double_t ch );
      virtual Double_t GetgVeff( Gfitter::GTypes::Particle particle, Double_t T3, Double_t ch );

   protected:
      
      // update of flags in case of parameter change
      void UpdateLocalFlags( Gfitter::GReference& ref );
   
      // update parameters
      void Update();
  
      // formfactors
      Double_t rhof  ( Gfitter::GTypes::Particle particle, Double_t Q );
      Double_t kappaf( Gfitter::GTypes::Particle particle, Double_t Q );
      
      // leading order contribution
      std::complex<Double_t> kappafLO( Gfitter::GTypes::Particle particle, Double_t ch );
      std::complex<Double_t> rhofLO  ( Gfitter::GTypes::Particle particle, Double_t ch );
      std::complex<Double_t> Uff     ( Gfitter::GTypes::Particle particle, Double_t ch );

      // partial widths of Z0
      Double_t GammaZff( Gfitter::GTypes::Particle particle, Double_t ch, Double_t mf = 0 );  

       // the dependents
      Gfitter::GParameterRef p_DeltaOMS;

      // reference to AlphasRunning
      GSM::RunningAlphaQCD& GetAlphasRun() const { return (GSM::RunningAlphaQCD&)*t_AlphasRun; }
      Gfitter::GTheoryRef    t_AlphasRun; 
      
      // reference to Vertex
      Vertex& GetVertex() const { return (Vertex&)*t_Vertex; }      
      Gfitter::GTheoryRef    t_Vertex; 
     
      // reference to QCDCorrections
      QCDCorrections& GetQCDCorr() const { return (QCDCorrections&)*t_QCDCorr; }   
      Gfitter::GTheoryRef    t_QCDCorr; 
     
      // reference to FermionPart
      FermionPart& GetFermionPart() const { return (FermionPart&)*t_FermionPart; } 
      Gfitter::GTheoryRef    t_FermionPart; 
     
      // reference to BosonPart
      BosonPart& GetBosonPart() const { return (BosonPart&)*t_BosonPart; }      
      Gfitter::GTheoryRef    t_BosonPart; 
      
      // reference to EW2Loop
      EW2Loop& GetEW2Loop() const { return (EW2Loop&)*t_EW2Loop; }      
      Gfitter::GTheoryRef    t_EW2Loop; 

      // reference to radiator function
      RadiatorFunctions& GetRadFun() const { return (RadiatorFunctions&)*t_radFun; }
      Gfitter::GTheoryRef    t_radFun;

      // reference to W mass
      Double_t GetMW() const { return ((GSM::WMass&)*t_MW).GetValue(); }      
      Gfitter::GTheoryRef t_MW; 

      // updated variables
      Double_t m_kapQCD;
      Double_t m_rhoQCD;
      Double_t m_R;
      Double_t m_CorrBB;
      Double_t m_alphaSTop;
      Double_t m_kapmix;

      Double_t m_Renorm;
      Double_t m_KapCorr;
      Double_t m_RhoTop;
      Double_t m_TBCorL;
      Double_t m_RhoCL;
      Double_t m_KapCL;

      Double_t m_RhoRemAdd;
      Double_t m_KapRemAdd;
      
      Double_t m_DRhoBar;
      Double_t m_DRhoBarLO;
      Double_t m_DrRem;

      // additional corrections for bottoms
      Double_t m_taub1;
      Double_t m_taub2;
      
      // non-factrorizable EW*QCD corrctions to Z->ff
      Double_t m_EWQCD[5];
      
      // bosonic + fermionic corr.
      std::complex<Double_t> m_ZAtMZ;
      std::complex<Double_t> m_WAtMW;
      std::complex<Double_t> m_ZFAtMZ;
      std::complex<Double_t> m_WAt0;
      std::complex<Double_t> m_MPhoZAtMZ;
     
      // OMS type
      enum OMSType { OMS1 = 1, OMS1_2, OMS2 };
      OMSType GetOMSType() const { return m_OMSType; }
      OMSType m_OMSType;
   
      Bool_t m_logMH;

   private:

      // boolean flags to track recomputation
      Bool_t m_isUpToDate_Update; 
          
      ClassDef(Z0ZFitter,0)
   };
}

//--------------------------- inline function for fast access ------------------------

// partial widths of Z->ff

inline Double_t GSM::Z0ZFitter::GetGaZelectron()
{
   return GammaZff( GTypes::kElectron, -1, GConstants::me() );
}

inline Double_t GSM::Z0ZFitter::GetGaZmuon()
{
   return GammaZff( GTypes::kMuon, -1, GConstants::mmu() );
}

inline Double_t GSM::Z0ZFitter::GetGaZtau()
{   
   return GammaZff( GTypes::kTau, -1, GConstants::mtau() );
}

inline Double_t GSM::Z0ZFitter::GetGaZneutrino()
{   
   return GammaZff( GTypes::kNeutrino, 0, 0);
}

inline Double_t GSM::Z0ZFitter::GetGaZup()
{   
   return GammaZff( GTypes::kUp, GMath::TwoThird() );
}

inline Double_t GSM::Z0ZFitter::GetGaZdown()
{   
   return GammaZff( GTypes::kDown, -GMath::OneThird() );
}

inline Double_t GSM::Z0ZFitter::GetGaZstrange()
{   
   return GammaZff( GTypes::kStrange, -GMath::OneThird() );
}

inline Double_t GSM::Z0ZFitter::GetGaZcharm()
{   
   return GammaZff( GTypes::kCharm, GMath::TwoThird() );
}

inline Double_t GSM::Z0ZFitter::GetGaZbottom()
{   
   return GammaZff( GTypes::kBottom, -GMath::OneThird() );
}

// effective couplings / A = axial, V = vector

inline Double_t GSM::Z0ZFitter::GetgAeff( Gfitter::GTypes::Particle particle, Double_t T3, Double_t ch )
{   
   Double_t gA = TMath::Sqrt(rhof(particle,ch))*T3;    
  
   return gA;
}    

inline Double_t GSM::Z0ZFitter::GetgVeff( Gfitter::GTypes::Particle particle, Double_t T3, Double_t ch )
{
   Double_t sineff = GetSin2Eff(particle);
   Double_t gV     = TMath::Sqrt(rhof(particle, ch))*(T3 - 2.0*ch*sineff) ;
         
   return gV;
}

#endif
