/**********************************************************************************
 * Package: GSM                                                                   *
 * Class  : ZMath                                                                 *
 *                                                                                *
 * Description:                                                                   *
 *      Auxillary Functions, for ZFITTER option                                   *
 *                                                                                *
 * Sources:                                                                       *
 *       - Z. Phys. C44, 493-502 (1989)                                           *
 *      - Bardin et al., ZFitter package dizet6_42.f                              *
 *                                                                                *
 * This class also contains code lines ported to C++ from the Fortran package     *
 * ZFITTER                                                                        *
 *                                                                                *   
 **********************************************************************************/

#include "TMath.h"
#include <iostream>

#include "GSM/ZMath.h"

using namespace std;

// reference Z. Phys. C44, 493-502 (1989)
// to all functions were added extra features
// to avoid divergences

// eq. (D.1) Z. Phys. C44, 493-502 (1989) 
// see dizet6_42.f line 1082-1107
std::complex<Double_t> GSM::ZMath::I0( const Double_t& MW2, const Double_t& Q2, const Double_t& m12, const Double_t& m22 )
{    
   std::complex<Double_t> i0(0,0);
	 

   if (m22/m12 < 1e-4) {
      Double_t RELQ = TMath::Log(TMath::Abs(1.0 + Q2/m12));
      if (-Q2 > Gfitter::GMath::IPow(sqrt(m12)+sqrt(m22),2)) {
         std::complex<Double_t> LQ (RELQ,-TMath::Pi());
         i0 = TMath::Log(m12/MW2) - 2.0 + (1.0 + m12/Q2)*LQ;
      }
      else i0 = TMath::Log(m12/MW2) - 2.0 + (1.0 + m12/Q2)*RELQ;
   }
   else {
      i0 = TMath::Log(m12/MW2) - 2.0 - (1.0/2.0 + (m12-m22)/(2.0*Q2))*TMath::Log(m12/m22) + 1.0/(2.0*Q2)*L(Q2,m12,m22);
   }
   return i0;
}

// eq. (D.2) Z. Phys. C44, 493-502 (1989)
// see dizet6_42.f line 1109-1136
std::complex<Double_t> GSM::ZMath::I1( const Double_t& MW2, const Double_t& Q2, const Double_t& m12, const Double_t& m22 )
{
   std::complex<Double_t> i1(0,0);
  
   if (m22/m12 < 1e-4) {

      Double_t RELQ = TMath::Log(TMath::Abs(1.0 + Q2/m12));
      if (-Q2 > Gfitter::GMath::IPow(sqrt(m12)+sqrt(m22),2)) {
         std::complex<Double_t> LQ (RELQ,-TMath::Pi());
         i1 = 0.5*TMath::Log(m12/MW2) - 1.0 - m12/(2.0*Q2) + (1.0 + m12/Q2)*1/2.0*LQ;
      }
      else i1 = 0.5*TMath::Log(m12/MW2) - 1.0 - m12/(2.0*Q2) + (1.0 + m12/Q2)*1/2.0*RELQ;
   }
   else {
      i1 = 0.5*TMath::Log(m12/MW2) - 1.0 - (m12-m22)/(2.0*Q2)
         - (1.0 + 2.0*m12/Q2 + Gfitter::GMath::IPow((m12-m22),2)/(Q2*Q2))*1/4.0*TMath::Log(m12/m22) 
         + (1.0 + (m12-m22)/Q2)*1.0/(4.0*Q2)*L(Q2,m12,m22);
   }
   return i1;
}

// eq. (D.3) Z. Phys. C44, 493-502 (1989) 
// see dizet6_42.f line 1138-1167
std::complex<Double_t> GSM::ZMath::I3( const Double_t& MW2, const Double_t& Q2, const Double_t& m12, const Double_t& m22 )
{
   std::complex<Double_t> i3(0,0);
   Double_t AL1W  = TMath::Log(m12/MW2);

   if (m22/m12 < 1e-4) {
      Double_t AQ = m12/Q2;
      Double_t RELQ = TMath::Log(TMath::Abs(1.0 + 1.0/AQ));
      if (-Q2 > Gfitter::GMath::IPow(sqrt(m12)+sqrt(m22),2)) {
         std::complex<Double_t> LQ (RELQ,-TMath::Pi());
         i3 = AL1W/6.0 - 5.0/18.0 + AQ/3.0 + AQ*AQ/3.0 + (1.0 - 2.0*AQ)*(1.0 + AQ)*(1.0 + AQ)*1.0/6.0*LQ;
      }
      else {
         i3 = ( 1.0/6.0*TMath::Log(m12/MW2) - 5.0/18.0 + m12/(3.0*Q2) + m12*m12/(3.0*Q2*Q2)
                + (1.0 + 2.0*m12/Q2)*Gfitter::GMath::IPow(1.0 + m12/Q2,2)*1.0/6.0*RELQ );
      }
   }
   else {
      i3 = ( 1.0/6.0*TMath::Log(m12/MW2) - 5.0/18.0 + (m12+m22)/(3.0*Q2) + 1/3.0*Gfitter::GMath::IPow((m12-m22)/Q2,2)
             + (-0.5 + 3.0/2.0*(m12+m22)*(m12-m22)/(Q2*Q2) + Gfitter::GMath::IPow((m12-m22)/Q2,3))*1/6.0*TMath::Log(m12/m22)
             + (0.5 - (m12+m22)/(2.0*Q2) - Gfitter::GMath::IPow((m12-m22)/Q2,2))*1.0/(6.0*Q2)*L(Q2,m12,m22) );
   }

   return i3;
}

// eq. (A.8) Z. Phys. C44, 493-502 (1989) 
// see dizet6_42.f line 1030-1048
std::complex<Double_t> GSM::ZMath::L( const Double_t& Q2, const Double_t& m12, const Double_t& m22 )
{    
   Double_t LAM = (Q2 + m12 + m22)*(Q2 + m12 + m22) - 4.0*m12*m22;
   std::complex<Double_t> REL = LAM*JJ(Q2,m12,m22);
   Double_t IL  = 0;
   if (-Q2 > Gfitter::GMath::IPow(sqrt(m12)+sqrt(m22),2) && LAM >= 0) IL = 2.0*TMath::Pi()*sqrt(LAM);
   std::complex<Double_t> l(real(REL),IL);
    
   return l;
}
  
// eq. (C.6) Z. Phys. C44, 493-502 (1989) 
// see dizet6_42.f line 1010-1028
std::complex<Double_t> GSM::ZMath::J( const Double_t& Q2, const Double_t& m12, const Double_t& m22 )
{    
   Double_t LAM = (Q2 + m12 + m22)*(Q2 + m12 + m22) - 4.0*m12*m22;
   std::complex<Double_t> REJ = JJ(Q2,m12,m22);
   Double_t IJ  = 0;
   if (-Q2 > Gfitter::GMath::IPow(sqrt(m12)+sqrt(m22),2)) IJ = 2.0*TMath::Pi()/sqrt(LAM);
   std::complex<Double_t> j (real(REJ),IJ);
    
   return j;
}

// Delta L
// see dizet6_42.f line 981-1008
std::complex<Double_t> GSM::ZMath::DL( const Double_t& Q2, const Double_t& Q2SBT, const Double_t& m12, const Double_t& m22 )
{
   std::complex<Double_t> dl(0,0);
   Double_t Q2S   = Q2SBT + m12 + m22;
   Double_t LAM   = Q2S*Q2S - 4.0*m12*m22;
   Double_t SLAM  = TMath::Sqrt(TMath::Abs(LAM));
   Double_t QD    = Q2 -Q2SBT;
   Double_t RQD   = TMath::Abs(QD/SLAM);
   if (RQD < 1e-3) {
      Double_t R = 4.0*m12/m22;
      if (R-1.0 != 0) {
         std::complex<Double_t> JS = J(Q2SBT,m12,m22);
         dl = ( 2.0 + Q2S*JS + QD/LAM*(Q2S - 2.0*m12*m22*JS) 
                + (QD/LAM)*(QD/LAM)*(-Q2S*Q2S/3.0 - 8.0/3.0*m12*m22*Q2S*JS) );
      }
      else dl = 4.0 + 2.0/3.0*QD/m22 - 2.0/15.0*QD*QD/(m22*m22);
   }
   else dl = (L(Q2,m12,m22) - L(Q2SBT,m12,m22))/QD;

   return dl;
}

// Auxiliary function for JJ  
// see dizet6_42.f line 1050-1080
std::complex<Double_t> GSM::ZMath::JJ( const Double_t& Q2, const Double_t& m12, const Double_t& m22 )
{    
   std::complex<Double_t> jj (0,0);
   Double_t Q2M  = (Q2 + m12 + m22);
   Double_t Q2L  =  4.0*m12*m22;
   Double_t LAM  = Q2M*Q2M - Q2L;
   Double_t R1   = 0;
   Double_t SLAM = 0;

   if (LAM < 0) {
      SLAM = sqrt(-LAM);
      if (Q2M < 0) {
         R1 = 2.0/SLAM*TMath::ATan(SLAM/Q2M);
         jj = R1 + 2.0*TMath::Pi()/SLAM;
      }
      if (Q2M == 0) {
         jj = 0.5*TMath::Pi()/sqrt(m12*m22);
      }
      if (Q2M > 0) {
         R1 = 2.0/SLAM*TMath::ATan(SLAM/Q2M);
         jj = R1;
      }
   }
   if (LAM ==0) {
      jj = 2.0/Q2M;
   }
   if (LAM > 0) {
      SLAM = sqrt(LAM);
      if (Q2M < 0) {
         jj = TMath::Log(Q2L/Gfitter::GMath::IPow((Q2M-SLAM),2))/SLAM;
      }
      if (Q2M >= 0) {
         jj = TMath::Log(Gfitter::GMath::IPow((Q2M+SLAM),2)/Q2L)/SLAM;
      }
   }
   return jj;
}

// Delta I0
// see dizet6_42.f line 1169-1179
std::complex<Double_t> GSM::ZMath::DI0( const Double_t& Q2, const Double_t& MV2, const Double_t& m12, const Double_t& m22 )
{
   std::complex<Double_t> di0(0,0);
   Double_t L12 = TMath::Log(m12/m22);
   di0 = (L(Q2, m12, m22) - (m12-m22)*L12)/(2.0*Q2) - DL(Q2, -MV2, m12, m22)/2.0;

   return di0;
}

// Delta I1
// see dizet6_42.f line 1181-1210
std::complex<Double_t> GSM::ZMath::DI1( const Double_t& Q2, const Double_t& MV2, 
                                          const Double_t& m12, const Double_t& m22, const Double_t& MW2 )
{    
   std::complex<Double_t> di1(0,0);
    
   if (m22 < 1e-10) {
      Double_t QV = (Q2 + MV2)/MV2;
      if (TMath::Abs(QV) < 1e-4) {
         Double_t R1V = m12/MV2;
         Double_t VQ  = MV2/Q2;
         Double_t AQ  = m12/Q2;
         di1 = 0.5*(-1.0 + R1V*(1.0-VQ) - 0.5*QV + (2.0*AQ + VQ*(VQ-1.0)*R1V*R1V)*TMath::Log(TMath::Abs(1.0+1.0/AQ)));
      }
      else di1 = (I1(MW2,Q2,m12,m22) - I1(MW2,-MV2,m12,m22))/QV;
   }
   else {
      Double_t AL12 = TMath::Log(m12/m22);
      Double_t Dm12 = (m12-m22)/Q2;
      Double_t AQ   = Q2/MV2;
      Double_t SMV1 = 1.0 -AQ;
      di1 = ( -Dm12/2.0 - (2.0*m12/Q2 + Dm12*Dm12*SMV1)*AL12/4.0
              + (1.0 + Dm12*SMV1)/(4.0*Q2)*L(Q2,m12,m22)
              - (1.0 - (m12-m22)/MV2)/4.0*DL(Q2,-MV2,m12,m22) );
   }
   return di1;
}

// Delta I3
// see dizet6_42.f line 1212-1247
std::complex<Double_t> GSM::ZMath::DI3( const Double_t& Q2, const Double_t& MV2, const Double_t& m12, 
                                          const Double_t& m22, const Double_t& MW2 )
{    
   std::complex<Double_t> di3(0,0);
    
   if (m22 < 1e-10) {
      Double_t QV = (Q2 + MV2)/MV2;
      if (TMath::Abs(QV) < 1e-4) {
         Double_t R1V = m12/MV2;
         Double_t VQ  = MV2/Q2;
         Double_t VQ2 = 1.0 - VQ + VQ*VQ;
         Double_t AQ  = m12/Q2;
         di3 = ( -1.0/6.0 + (VQ-0.5)/3.0*R1V + VQ2/3.0*R1V*R1V
                 -QV*(0.5+R1V)/6.0 - VQ*R1V*R1V*((VQ-1.0)/2.0 + VQ2/3.0*R1V)*TMath::Log(TMath::Abs(1.0+1.0/AQ)) );
      }
      else di3 = (I3(MW2,Q2,m12,m22) - I1(MW2,-MV2,m12,m22))/QV;
   }
   else {
      Double_t AL12 = TMath::Log(m12/m22);
      Double_t Dm12 = (m12-m22)/Q2;
      Double_t Sm12 = (m12+m22)/Q2;
      Double_t AQ   = Q2/MV2;
      Double_t SMV1 = 1.0 - AQ;
      Double_t SMV2 = 1.0 - AQ + AQ*AQ;
      di3 = ( Sm12/3.0 + Dm12*Dm12/3.0*SMV1
              + (Sm12*Dm12/4.0*SMV1 + Gfitter::GMath::IPow(Dm12,3)/6.0*SMV2)*AL12
              + (0.5 - Sm12/2.0*SMV1 - Dm12*Dm12*SMV2)/(6.0*Q2)*L(Q2,m12,m22)
              - (0.5 + 0.5*(m12+m22)/MV2
                 - Gfitter::GMath::IPow((m12-m22)/MV2,2))/6.0*DL(Q2,-MV2,m12,m22) ); 
   }
   return di3;
}

// ------- mtop^4 contribution from Barbieri et. al. ---------      
// ------- Nucl. Phys. B409(1993) 105-107 --------------------

// see dizet6_42.f line 4547-4563
Double_t GSM::ZMath::BarbMc( const Double_t& x )
{  
   Double_t P[8] = {-0.74141, -11.483, 9.6577, -6.727, 3.0659, -0.82053, 0.11659, -0.67712e-2};
   Double_t Barb = 0;
   if (x < 4) {
      for (Int_t i = 0; i < 8; i++) {  Barb += P[i]*Gfitter::GMath::IPow(x,i); }
   }
   else {
      Double_t Pi2      = Gfitter::GMath::IPow(TMath::Pi(),2);
      Double_t Inv2     = 1/(x*x);
      Double_t LogInv2  = TMath::Log(Inv2);
      Barb              = ( 49/4.0 + Pi2 + 27/2.0*LogInv2 + 3/2.0*LogInv2*LogInv2
                            + Inv2/3.0*(2.0 - 12.0*Pi2 + 12.0*LogInv2 -27.0*LogInv2*LogInv2)
                            + Inv2*Inv2/48.0*(1613.0 - 240.0*Pi2 - 1500.0*LogInv2 - 720.0*LogInv2*LogInv2) );
   }
   return Barb;
}
      
// see dizet6_42.f line 4565-4583
Double_t GSM::ZMath::BarbMb( const Double_t& x )
{  
   Double_t P[8] = {5.6807, -11.015, 12.814, -9.2954, 4.3305, -1.2125, 0.18402, -0.11582e-1};
   Double_t Barb = 0;
   if (x < 4) {
      for (Int_t i = 0; i < 8; i++) {  Barb += P[i]*Gfitter::GMath::IPow(x,i); }
   }
   else {
      Double_t Pi2      = Gfitter::GMath::IPow(TMath::Pi(),2);
      Double_t Inv2     = 1/(x*x);
      Double_t LogInv2  = TMath::Log(Inv2);
      Barb              = 1/144.0*( 311.0 + 24.0*Pi2 + 282.0*LogInv2 + 90.0*LogInv2*LogInv2
                                    - 4.0*Inv2*(40.0 + 6.0*Pi2 + 15.0*LogInv2 + 18.0*LogInv2*LogInv2)
                                    + 3.0*Inv2*Inv2*(242.09 - 60.0*Pi2 - 454.2*LogInv2 - 180.0*LogInv2*LogInv2) );
   }
   return Barb;
}
      
// Supplies real parts of J0,S3,S30 (w,t) in analytic presentation
// S3Wana(mt2,MW2,-s,...) supplies `w' indices
// S3Wana(MW2,mt2,-s,...) supplies `t' indices
// see dizet6_42.f line 4063-4093
void GSM::ZMath::S3Wana ( const Double_t& mt2, const Double_t& MW2, const Double_t& s,
                            Double_t& J0, Double_t& S3, Double_t& S30 )
{
   std::complex<Double_t> Cmt2 (mt2, -1e-10);
   std::complex<Double_t> CMW2 (MW2, -1e-10);
   std::complex<Double_t> Sqrt = std::sqrt( 1.0 + 4.0*CMW2/s );
   std::complex<Double_t> LogS = ZMath::Log( (Sqrt + 1.0)/(Sqrt - 1.0) );
	
   std::complex<Double_t> x0   = (Cmt2 - CMW2)/s;
   std::complex<Double_t> x1   = 0.5*(1.0 - Sqrt);
   std::complex<Double_t> x2   = 0.5*(1.0 + Sqrt);
   std::complex<Double_t> x3   = Cmt2/(Cmt2 - CMW2);

   std::complex<Double_t> y1   = x1/x0;
   std::complex<Double_t> y2   = (1.0 - x1)/(1.0 - x0);
   std::complex<Double_t> y3   = x2/x0;
   std::complex<Double_t> y4   = (1.0 - x2)/(1.0 - x0);
   std::complex<Double_t> y5   = x3/x0;
   std::complex<Double_t> y6   = (1.0 - x3)/(1.0 - x0);
	 
   // results
   J0                    = real( Sqrt*LogS );
   if (mt2 >= MW2) S30 = real( 1.0/s*LogS*LogS );
   else            S30 = real( 1.0/s*(Gfitter::GMath::Zeta2() - ZMath::Li2(1.0 - s/mt2)) );
   S3                  = real( 1.0/s*( + ZMath::Li2( 1.0/(1.0-y1) ) - ZMath::Li2( 1.0/(1.0-y2) ) 
                                       + ZMath::Li2( 1.0/(1.0-y3) ) - ZMath::Li2( 1.0/(1.0-y4) ) 
                                       - ZMath::Li2( 1.0/(1.0-y5) ) + ZMath::Li2( 1.0/(1.0-y6) ) ) );
	 
}

// see dizet6_42.f line 5484-5519
Double_t GSM::ZMath::Afmt3( const Double_t& alst, const Double_t& mt2, const Double_t& MZ2, const Double_t& sW2 )
{
   Double_t TS2   =  0.2604341376322; 
   Double_t TD3   = -3.0270094939877;
   Double_t TB4   = -1.7628000870738;
   Double_t LogZT =  TMath::Log(MZ2/mt2);
   Int_t nf       =  6;

   Double_t CA1   = -2/3.0*(1.0 + 2*Gfitter::GMath::Zeta2());
   Double_t CA2C  = ( MZ2/mt2*(-17.224 + 0.08829*LogZT + 0.4722*LogZT*LogZT
                               + ( 22.6367 + 1.2527*LogZT - 0.8519*LogZT*LogZT)*sW2) );
   Double_t CA2L  = ( 157/648.0 - 3313/162.0*Gfitter::GMath::Zeta2() - 308/27.0*Gfitter::GMath::Zeta3()
                      + 143/18.0*Gfitter::GMath::Zeta4() - 4/3.0*Gfitter::GMath::Zeta2()*TMath::Log(2.0)
                      + 441/8.0*TS2 - 1/9.0*TB4 - 1/18.0*TD3 
                      - (1/18.0 - 13/9.0*Gfitter::GMath::Zeta2() + 4/9.0*Gfitter::GMath::Zeta3())*nf );
   // - (11/6.0 - 1/9.0*nf)*(1.0 + 2*Gfitter::GMath::Zeta2())*TMath::Log(mt2/mt2)
   Double_t CA2I  = ( Gfitter::GMath::IPow(MZ2/mt2,2)*
                      (-7.7781 - 0.072263*LogZT + 0.004938*LogZT*LogZT
                       + (21.497 + 0.05794*LogZT - 0.006584*LogZT*LogZT)*sW2 - 21.0799*sW2*sW2) );

   Double_t CA2   = CA2L + CA2C + CA2I;
	  
   return CA1*alst/TMath::Pi() + CA2*Gfitter::GMath::IPow( alst/TMath::Pi() ,2 );
}

// see dizet6_42.f line 5521-5545
Double_t GSM::ZMath::TbQCD( const Double_t& alst, const Double_t& mt2, const Double_t& MZ2, const Double_t& sW2 )
{
   Double_t cW2   = 1.0 - sW2;
   Double_t logzt = TMath::Log( MZ2/mt2);
  
   Double_t Ca2c  = ( MZ2/mt2*( +(-11.3184 - 0.62630*logzt + 0.4259*logzt*logzt)*sW2
                                +(+22.6367 + 1.25270*logzt - 0.8519*logzt*logzt)*sW2) );
   Double_t Ca2i  = ( GMath::IPow( (MZ2/mt2), 2 )*( (-16.01860 - 0.02897*logzt + 0.003292*logzt*logzt)*sW2 
                                                    + 10.54*sW2*sW2 
                                                    + (21.497 + 0.05794*logzt - 0.006584*logzt*logzt)*sW2 - 21.0799*sW2*sW2) );
 
   return -cW2/sW2*(Ca2c + Ca2i)*GMath::IPow( (alst/TMath::Pi()), 2 );
}


// ------------ some polarization operators and their auxiliary functions ----------------

// see bcqcd5_14.f line 289-298
complex<Double_t> GSM::ZMath::PolVfi( const Double_t& al )
{
   return ( 45.0/(4.0*al) - 82.0/(27.0*al*al) + 449.0/(900.0*al*al*al)
            - 62479.0/(661500.0*al*al*al*al) );
}
   
// see bcqcd5_14.f line 390-399
complex<Double_t> GSM::ZMath::PolAfi( const Double_t& al )
{    
   return ( -93.0/2.0+67.0/12.0/al-689.0/540.0/(al*al)
            +1691.0/12600.0/(al*al*al)-1.8599e-2/(al*al*al*al) );
}      

// see bcqcd5_14.f line 340-349
complex<Double_t> GSM::ZMath::DPolVfi( const Double_t& al )
{
   return ( 45.0/4.0 - 164.0/(27.0*al) + 449.0/(300.0*al*al)
            - 62479.0/(165375.0*al*al*al) + 9.55063e-2/(al*al*al*al) );
}
  
// see bcqcd5_14.f line 442-451
complex<Double_t> GSM::ZMath::DPolAfi( const Double_t& al )
{
   return ( 67.0/12.0-689.0/270.0/al+1691.0/4200.0/(al*al)
            -7.43961e-2/(al*al*al)+275205680132.0/18606865047887.0/(al*al*al*al) );
}

	
// Li_2(y + i*epsilon*sign(Re(e))), see for example
// "The Standard Model in the Making" by Bardin and Passarino
std::complex<Double_t> GSM::ZMath::Li2( const std::complex<Double_t>& y, std::complex<Double_t> e )
{
   std::complex<Double_t> li2  (0,0);
   std::complex<Double_t> cone (1.0,0);
   std::complex<Double_t> x = y;
   Double_t B[59];
   ZMath::GetCoefficients( B );
	
   const Int_t N = 30;
	
   Int_t flag1 = 0;
   Int_t flag2 = 0;
	   
   if      (x == li2)  li2 = 0.0;
   else if (x == cone) li2 = GMath::Zeta2();
   else {
      if (abs(x) > 1.0) {
         x = cone/x;
         flag1 = 1;}
      if (real(x) > 0.5) {
         x = cone - x;
         flag2 = 1;}
      std::complex<Double_t> z = -ZMath::Log(cone-x);   
      std::complex<Double_t> h = cone;
      std::complex<Double_t> Cold (0.0,0.0);
      for (Int_t i=1; i<=N; i++) {
         Cold = li2;
         h    = h*z/(1.0*i);
         li2  = li2 + B[i-1]*h;
         if (B[i-1] != 0 && abs(Cold/li2-cone) < 1e-20) break;
      }
	
      if (flag2 == 1) li2 = -li2 + GMath::Zeta2() - ZMath::Log(x)*ZMath::Log(cone-x);
      if (flag1 == 1) li2 = -li2 - GMath::Zeta2() - 0.5*pow(CLog(-y,-e),2);
   }
	
   return li2;
}
	 
// Li_3(y + i*epsilon*sign(Re(e)))
std::complex<Double_t> GSM::ZMath::Li3( const std::complex<Double_t>& y, std::complex<Double_t> e )
{
   std::complex<Double_t> li3  (0,0);
   std::complex<Double_t> cone (1.0,0);
   std::complex<Double_t> x = y;
   std::complex<Double_t> h (0,0);

   Int_t flag = 0;
	
   if      (x == li3)  li3 = 0.0;
   else if (x == cone) li3 = GMath::Zeta3();
   else {
      if (abs(x) > 1.0) {
         x = cone/x;
         flag = 1;
      }
	
      if (real(x) > 0.5) {
         h   = ZMath::Log(x);
         li3 = ( - Li3H(cone-cone/x) - Li3H(cone-x) + GMath::Zeta3() + GMath::Zeta2()*h
                 + h*h*h/6.0 - 0.5*h*h*ZMath::Log(cone-x) );
      }
      else li3 = Li3H(x);
	
      if (flag == 1) {
         h = CLog(-y,-e);
         li3 = li3 - GMath::Zeta2()*h - h*h*h/6.0;
      }
   }
	
   return li3;
}
	
// Li_3(x) for 0 < |x| <= 1, Re(x) <= 1/2           
std::complex<Double_t> GSM::ZMath::Li3H( const std::complex<Double_t>& x )
{
   std::complex<Double_t> li3h (0,0);   
   std::complex<Double_t> cone (1.0,0);
   std::complex<Double_t> Cold (0,0);
   std::complex<Double_t> c    (0,0);
   std::complex<Double_t> z = -ZMath::Log(cone-x);
   std::complex<Double_t> p = cone;             
   Double_t B[59];
   ZMath::GetCoefficients( B );
	
   const Int_t N = 30;
   Double_t F[N+1];
   F[0] = 1.0;
   for (Int_t i=1; i <= N ; i++) F[i] = F[i-1]*1.0*i;
	   
   for (Int_t i=1; i <= N ; i++) {
      Cold = li3h;
      p    = p*z;
      c    = 0.0;
      for (Int_t j = 1; j <= i; j++) c = c + B[j-1]*B[i-j]/(F[j]*F[i-j]);
	
      li3h = li3h + p*c/(1.0*i);
      if (abs(Cold/li3h - cone) < 1e-20) break;
   }
	
   return li3h;
}
	 
// sqrt(x + i*epsilon*sign(Re(e)))
std::complex<Double_t> GSM::ZMath::CSqrt( const std::complex<Double_t>& x, const std::complex<Double_t>& e )
{
   std::complex<Double_t> crt(0.0,0.0);
   std::complex<Double_t> I(0.0,1.0);
	
   Double_t sign;
   if (real(e) < 0) sign = -1.0;
   else             sign = 1.0;
	
   if (real(x) < 0 && imag(x) == 0) crt = sign*I*sqrt(-x);
   else                             crt = sqrt(x);
	
   return crt;
}
	
// ln(x + i*epsilon*sign(Re(e)))
std::complex<Double_t> GSM::ZMath::CLog( const std::complex<Double_t>& x, const std::complex<Double_t>& e )
{
   std::complex<Double_t> cln(0.0,0.0);
   std::complex<Double_t> I(0.0,1.0);
	         
   Double_t sign;
   if (real(e) < 0) sign = -1.0;
   else             sign = 1.0;
	     
   if (real(x) < 0 && imag(x) == 0) cln = ZMath::Log(-x) + sign*I*TMath::Pi();
   else                             cln = ZMath::Log(x);
	   
   return cln;
}
	 
// Coefficients for Euler polylogarithm Li_n , see for example
// "The Standard Model in the Making" by Bardin and Passarino
void GSM::ZMath::GetCoefficients( Double_t Coef[59] )
{
   for (Int_t i=0; i<59; i++) Coef[i] = 0;
	   
   Coef[0]  = 1.0;
   Coef[1]  = -.50;
   Coef[2]  = 1.0/6.0;
   Coef[4]  = -1.0/30.0;
   Coef[6]  = 1.0/42.0;
   Coef[8]  = Coef[4];
   Coef[10] = 5.0/66.0;
   Coef[12] = -691.0/2730.0;
   Coef[14] = 7.0/6.0;
   Coef[16] = -3617.0/510.0;
   Coef[18] = 43867.0/798.0;
   Coef[20] = -174611.0/330.0;
   Coef[22] = 854513.0/138.0;
   Coef[24] = -236364091.0/2730.0;
   Coef[26] = 8553103.0/6.0;
   Coef[28] = -23749461029.0/870.0;
   Coef[30] = 8615841276005.0/14322.0;
   Coef[32] = -7709321041217.0/510.0;
   Coef[34] = 2577687858367.0/6.0;
   Coef[36] = -26315271553053477373.0/1919190.0;
   Coef[38] = 2929993913841559.0/6.0;
   Coef[40] = -261082718496449122051.0/13530.0;
   Coef[42] = 1520097643918070802691.0/1806.0;
   Coef[44] = -27833269579301024235023.0/690.0;
   Coef[46] = 596451111593912163277961.0/282.0;
   Coef[48] = -5609403368997817686249127547.0/46410.0;
   Coef[50] = 495057205241079648212477525.0/66.0;
   Coef[52] = -801165718135489957347924991853.0/1590.0;
   Coef[54] = 29149963634884862421418123812691.0/798.0;
   Coef[56] = -2479392929313226753685415739663229.0/870.0;
   Coef[58] = 84483613348880041862046775994036021.0/354.0;
}
