/**********************************************************************************
 * Package: GSM                                                                   *
 * Class  : ZMath                                                                 *
 *                                                                                *
 * Description:                                                                   *
 *      Namespace of                                                              *
 *      Auxillary Functions, particulary for ZFITTER option                       *
 *                                                                                *
 **********************************************************************************/

#ifndef GSM_ZMath
#define GSM_ZMath

#include <assert.h>
#include <complex>
#include "Rtypes.h"
#include "TMath.h"

#include "Gfitter/GMath.h"

using namespace Gfitter;

namespace GSM {
   
   namespace ZMath {
      
      // Infrared Integrals for one loop core of ZFITTER option
      std::complex<Double_t> I0 ( const Double_t& MW2, const Double_t& Q2, const Double_t& m12, const Double_t& m22 );
      std::complex<Double_t> I1 ( const Double_t& MW2, const Double_t& Q2, const Double_t& m12, const Double_t& m22 );
      std::complex<Double_t> I3 ( const Double_t& MW2, const Double_t& Q2, const Double_t& m12, const Double_t& m22 );
	 
      // Auxiliary functions for one loop core L and J
      std::complex<Double_t> L  ( const Double_t& Q2, const Double_t& m12, const Double_t& m22 );
      std::complex<Double_t> J  ( const Double_t& Q2, const Double_t& m12, const Double_t& m22 );
    
      // Auxiliary functions for J - don't use it outside
      std::complex<Double_t> JJ ( const Double_t& Q2, const Double_t& m12, const Double_t& m22 );
      
      // Delta I0, I1, I3
      std::complex<Double_t> DI0( const Double_t& Q2, const Double_t& MV2, const Double_t& m12, const Double_t& m22 );
      std::complex<Double_t> DI1( const Double_t& Q2, const Double_t& MV2, const Double_t& m12, const Double_t& m22, const Double_t& MW2 );
      std::complex<Double_t> DI3( const Double_t& Q2, const Double_t& MV2, const Double_t& m12, const Double_t& m22, const Double_t& MW2 );
      std::complex<Double_t> DL ( const Double_t& Q2, const Double_t& Q2SBT, const Double_t& m12, const Double_t& m22 );

      // function for top Vertex correction 
      // Supplies real parts of J0,S3,S30 (w,t) in analytic presentation
      // S3Wana(mt2,MW2,-s,...) supplies `w' indices
      // S3Wana(MW2,mt2,-s,...) supplies `t' indices
      void S3Wana ( const Double_t& mt2, const Double_t& MW2, const Double_t& s, Double_t& J0 , Double_t& S3 , Double_t& S30); 
 
      Double_t BarbMc ( const Double_t& x );
      Double_t BarbMb ( const Double_t& x );
      
      Double_t Afmt3  ( const Double_t& alst, const Double_t& mt2, const Double_t& MZ2, const Double_t& sW2 );
      Double_t TbQCD  ( const Double_t& alst, const Double_t& mt2, const Double_t& MZ2, const Double_t& sW2 );

      // math functions for ew 2 loop calculation
      Double_t J1    ( const Double_t& x , const Double_t& y );
      Double_t Lambda( const Double_t& d );
      Double_t Phi   ( const Double_t& x );
      Double_t revf  ( const Double_t& x );
      Double_t ug    ( const Double_t& x );
      Double_t gdeg  ( const Double_t& x );
      Double_t B0    ( const Double_t& q, const Double_t& m1, const Double_t& m2, const Double_t& mu2 );
      Double_t Omega ( const Double_t& a, const Double_t& b );
      Double_t hbar  ( const Double_t& x );
 
      // some polarization operators and their auxiliary functions
      std::complex<Double_t> PolVfi( const Double_t& al );
      std::complex<Double_t> PolAfi( const Double_t& al );
      std::complex<Double_t> DPolVfi( const Double_t& al );
      std::complex<Double_t> DPolAfi( const Double_t& al );
      
      // standard functions
      std::complex<Double_t> Log( const std::complex<Double_t>& z );
	
      // sqrt(x + i*epsilon*sign(Re(e)))
      std::complex<Double_t> CSqrt( const std::complex<Double_t>& x, const std::complex<Double_t>& e );
 
      // ln(x + i*epsilon*sign(Re(e)))
      std::complex<Double_t> CLog( const std::complex<Double_t>& x, const std::complex<Double_t>& e );
	     
      // polylogarithm Li_n
	     
      // Spence function = Li2 function = Int(ln(1+y)/y,y,0,x)
      std::complex<Double_t> Li2( const std::complex<Double_t>& y, std::complex<Double_t> e = 1.0);
	
      // Li3 function
      std::complex<Double_t> Li3( const std::complex<Double_t>& y, std::complex<Double_t> e = 1.0);
	     
      // Li_3(x) for 0 < |x| <= 1, Re(x) <= 1/2     
      std::complex<Double_t> Li3H( const std::complex<Double_t>& x );
	     
      // Clausen Integral
      Double_t Clausen( const Double_t& x );
	
      // Coefficients for Euler polylogarithm Li_n
      void GetCoefficients( Double_t Coef[59] );

   }
}

// ----------------- inline functions for fast access -----------------------------
// just the shorter implementations

//------------------------------ EW2loop-------------------------------------------
// see for all functions The Standard Model in the Making, Oxford 99 page 433-434


// eq. (A5) from Nucl. Phys. B248 (1948) 589-614
inline Double_t GSM::ZMath::J1( const Double_t& x , const Double_t& y )
{
   Double_t d =  4.0*x*y - 1.0;
   return 1/6.0*(TMath::Log(x) + 1/3.0 + (3.0 + d)*(Lambda(d) - 1.0));
}

// eq. (8b) from hep-ph/9603374, hep-ph/9905472 page 19 etc. 
inline Double_t GSM::ZMath::Lambda( const Double_t& d )
{
  Double_t ln   = 0;
  if( d < 0){
    Double_t d1 = TMath::Sqrt( -d );
    ln          = 0.5*d1*TMath::Log(TMath::Abs((1.0+d1)/(1.0-d1)));
  }
  else ln       = TMath::Sqrt(d)*TMath::ATan(1.0/TMath::Sqrt(d));

  return ln;
}

// eq. (8c) from hep-ph/9603374, hep-ph/9905472 page 19 etc. 
inline Double_t GSM::ZMath::Phi( const Double_t& x )
{
   Double_t phi    = 0;
   if (x < 1) phi = 4.0*TMath::Sqrt(x/(1.0-x))*ZMath::Clausen( 2.0*TMath::ASin( TMath::Sqrt(x) ) );
   else {
      phi = ( TMath::Sqrt(x/(x-1.0))*
              ( - 4.0*TMath::DiLog( 0.5*(1.0 - TMath::Sqrt(1.0 - 1.0/x)) ) 
                + 2.0*GMath::IPow( TMath::Log( 0.5*(1.0-TMath::Sqrt(1.0 - 1.0/x)) ), 2 )
                - GMath::IPow( TMath::Log(4.0*x) , 2 ) + GMath::IPow( TMath::Pi(), 2 )/3.0 ) );
   }
   return phi;
}

// (8.80) The Satndard Model in the Making
inline Double_t GSM::ZMath::revf( const Double_t& x )
{
   Double_t a1 = 2.0*(2.0 + x) - 0.5 + (3.0 + 2.0*x)*TMath::Log(x);
   Double_t a2 = ( GMath::IPow( 1.0 + x ,2 )*
                   ( GMath::IPow( TMath::Log(1.0 + 1.0/x) , 2 ) - GMath::IPow( TMath::Pi(), 2 )/3.0
                     + 2.0*TMath::DiLog( x/(x + 1.0) ) ) );
   return a1 + a2;
}

// (8.80) The Satndard Model in the Making
inline Double_t GSM::ZMath::ug( const Double_t& x )
{
   Double_t a1  = 4.0*x - 1.0;
   Double_t dat = 0;

   if (a1 < 0) {
      dat = 0.5*TMath::Log( TMath::Abs( 1.0 + TMath::Sqrt( TMath::Abs(a1) ) )/
                            TMath::Abs( 1.0 + TMath::Sqrt( TMath::Abs(a1) ) ) );
   }
   else dat = TMath::ATan( 1.0/TMath::Sqrt(a1) );

   return 4.0*x*(2.0 + x)*dat*dat - 1.5 + x - (1.0 + 2.0*x)*TMath::Sqrt( TMath::Abs(a1) )*dat; 
}

// (8.81) The Standard Model in the Making
// eq. (8a) from hep-ph/9603374, hep-ph/9905472 page 19 etc.
inline Double_t GSM::ZMath::gdeg( const Double_t& x )
{
   Double_t g = 0;
  
   if (x > 4) {
      g = 2.0*TMath::Sqrt(x/4.0 - 1.0)*TMath::Log( (1.0 - TMath::Sqrt(1.0-4.0/x))/
                                                   (1.0 + TMath::Sqrt(1.0-4.0/x)) );
   }
   else g = TMath::Sqrt(4.0 - x)*(TMath::Pi() - 2.0*TMath::ASin( TMath::Sqrt(x/4.0) ));

   return g;
}

// eq. (A2) from Phys. Rev. VOlume 46, Number 7 1992 
inline Double_t GSM::ZMath::B0( const Double_t& q, const Double_t& m1, const Double_t& m2, const Double_t& mu2 )
{
   return -( TMath::Log(q/mu2) - 2.0 
             + 0.5*( 1.0 + (m1/q - m2/q))*TMath::Log(m1/q) 
             + 0.5*( 1.0 - (m1/q - m2/q))*TMath::Log(m2/q)  
             + 2.0*Omega(m1/q, m2/q) );
}

// eq. (A3) from Phys. Rev. VOlume 46, Number 7 1992 
inline Double_t GSM::ZMath::Omega( const Double_t& a, const Double_t& b )
{
   Double_t C     = 0.5*(a+b) - 0.25*GMath::IPow( a-b, 2 ) - 0.25;
   Double_t omega = 0; 

   if (C > 0) {
      omega = ( TMath::Sqrt(C)*
                (TMath::ATan( (1.0 + a - b)/(2.0*TMath::Sqrt(C)) ) +
                 TMath::ATan( (1.0 - a + b)/(2.0*TMath::Sqrt(C)) )) );
   }
   else {
      C     = -C;
      omega =  ( 0.5*TMath::Sqrt(C)*
                 TMath::Log( ( 0.5*( a + b - 1.0) - TMath::Sqrt(C) )/
                             ( 0.5*( a + b - 1.0) + TMath::Sqrt(C) ) ) );
   }

   return omega;
}

// (8.76) The Standard Model in the Making
inline Double_t GSM::ZMath::hbar( const Double_t& x )
{
   Double_t h = 31/18.0 - x + x*x/3.0 + (1.0 - 1.5*x + 0.75*x*x - x*x*x/6.0)*TMath::Log(x);
   Double_t c = 0;

   if (x > 4) {
      Double_t a = TMath::Sqrt( 1.0 - 4.0/x );
      c = TMath::Sqrt( x*x/4.0 - x)*TMath::Log( (1.0 - a)/(1.0 + a) );
   }
   else {
      c = TMath::Sqrt( 4.0*x - x*x )*TMath::ATan( TMath::Sqrt( 4.0/x - 1.0 ) );
   }
 
   return h + (-2.0 + 2.0/(4.0 - x) + 5.0/6.0*x - x*x/3.0)*c;
}

inline std::complex<Double_t> GSM::ZMath::Log( const std::complex<Double_t>& x )
{
   return log(x);
}

inline Double_t GSM::ZMath::Clausen( const Double_t& x )
{
   std::complex<Double_t> I (0.0,1.0);
   return real( 0.5*I*( Li2( exp(-I*x) ) - Li2( exp(I*x) ) ) );
}

#endif
