/**********************************************************************************
 * Project: GSTU - STU Parameter fitting package                                  *
 * Package: GSTU                                                                  *
 * Class  : RandallSundrum_CustSym                                                *
 *                                                                                *
 * Description:                                                                   *
 *      implementation                                                            *
 *                                                                                *
 * Sources:                                                                       *
 *      S. Casagrande, F. Goertz, U. Haisch, M. Neubert, T. Pfoh                  * 
 *            Flavor Physics in the Randall-Sundrum Model:                        *
 *            I. Theoretical Setup and Electroweak Precision Tests                *
 *            JHEP 0810, 094 (2008)                                               *
 *            arXiv:0807.4937                                                     * 
 *                                                                                *
 *      K. Agashe, A. Delgado, M.J. May, and R. Sundrum                           *
 *            JHEP 0308, 050 (2003)                                               *
 *            [hep-ph/030836]                                                     *
 *                                                                                *
 * see corresponding .h file for author and license information                   *         
 *                                                                                *
 **********************************************************************************/

#include "TMath.h"

#include "Gfitter/GTheory.h"
#include "Gfitter/GTheoryRef.h"
#include "Gfitter/GParameterRef.h"
#include "Gfitter/GConstants.h"

#include "GSTU/RandallSundrum_CustSym.h"

using namespace Gfitter;

ClassImp(GSTU::RandallSundrum_CustSym)

GSTU::RandallSundrum_CustSym::RandallSundrum_CustSym()
   : TheoryBase()
{
   SetTheoryName( GetName() );
   SetExistDerivative( kFALSE );

   // book SM parameters
   BookTheory   ( "GSM::MW"   ,&t_MW );
   BookParameter( "MZ"        ,&p_MZ  );
   
   // book RS parameters
   BookParameter( "MKK" ,&p_MKK ); 
   BookParameter( "L"   ,&p_L   ); 
     
}


void GSTU::RandallSundrum_CustSym::Initialise(){}

// equations refer to JHEP 0810, 094 (2008) 
// eq. (153)
Double_t GSTU::RandallSundrum_CustSym::GetS()
{

   Double_t vev = GConstants::vev();
   Double_t pi  = TMath::Pi();
   
   Double_t s   = 2.*pi*vev*vev/(p_MKK*p_MKK)*(1. - 1./p_L);

   return s;
}

// eq. (153)
Double_t GSTU::RandallSundrum_CustSym::GetT()
{
   Double_t vev = GConstants::vev();
   Double_t pi  = TMath::Pi();
   Double_t MW  = GetMW().GetTheoryPrediction();
   Double_t cw2 = TMath::Power( MW/p_MZ, 2 );
   Double_t t   = -pi*vev*vev/(4.*cw2*p_MKK*p_MKK)*(1./p_L);

   return t;
}
